@@ -2,12 +2,9 @@
  * buffer.c - Packet buffering code.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
- *
- * FIXME: This stuff should probably record the protocol that the
- * packet came in over so that we can use that info when forwarding
- * the packet. (It may not always be coming in over IP!)
  */
 
 #include "diald.h"
@@ -47,7 +44,8 @@ void buffer_packet(unsigned int len,unsigned char *pkt)
 
     buffer_check();
     if (len+6 > buffer_size) {
-	syslog(LOG_ERR,"Can't buffer packet of length %d, buffer to small.",
+	mon_syslog(LOG_NOTICE,
+	    "Can't buffer packet of length %d, buffer too small",
 	    len);
 	return;
     }
@@ -82,7 +80,7 @@ void buffer_packet(unsigned int len,unsigned char *pkt)
 	    tail = (tail+1)%buffer_size;
 	}
     } else {
-	syslog(LOG_ERR,
+	mon_syslog(LOG_NOTICE,
 	    "Can't buffer packet of length %d, only %d bytes available.",
 	    len,buffer_size-(used+6));
     }
@@ -93,37 +91,68 @@ void forward_buffer()
     unsigned int clen, i;
     unsigned long stamp;
     unsigned long ctime = timestamp();
-    struct SOCKADDR to;
+    struct sockaddr_pkt sp;
+#ifdef HAVE_AF_PACKET
+    struct sockaddr_ll sl;
+#endif
+    struct sockaddr *to;
+    size_t to_len;
     unsigned char pkt[4096];
 
     buffer_check();
 
-    memset(&to,0,sizeof(to));
-#ifdef HAS_SOCKADDR_PKT
-    to.spkt_family = AF_INET;
-    strcpy(to.spkt_device,snoop_dev);
-    to.spkt_protocol = htons(ETH_P_IP);
-#else
-    to.sa_family = AF_INET;
-    strcpy(to.sa_data,snoop_dev);
+#ifdef HAVE_AF_PACKET
+    if (af_packet) {
+	memset(&sl, 0, sizeof(sl));
+	sl.sll_family = AF_PACKET;
+	sl.sll_protocol = htons(ETH_P_IP);
+	sl.sll_ifindex = snoop_index;
+	to = (struct sockaddr *)&sl;
+	to_len = sizeof(sl);
+    } else
 #endif
+    {
+	memset(&sp, 0, sizeof(sp));
+	sp.spkt_family = AF_INET;
+	strcpy(sp.spkt_device, snoop_dev);
+	sp.spkt_protocol = htons(ETH_P_IP);
+	to = (struct sockaddr *)&sp;
+	to_len = sizeof(sp);
+    }
 
     while (used > 0) {
 	clen = (B(head)<<8) | B(head+1);
 	stamp = (B(head+2)<<24) | (B(head+3)<<16) | (B(head+4)<<8) | B(head+5);
         if (stamp+buffer_timeout >= ctime) {
+	    unsigned short wprot;
+	    unsigned char *dpkt;
+	    unsigned int dlen;
+
 	    for (i = 0; i < clen; i++)
 		pkt[i] = B(head+6+i);
-	    /* Subtle point here: normally we forward packets on the
-	     *	"fwdfd" socket pointer. If we do this here, then
-	     * the packet we foward will get seen by the filter again.
-             * Since we've already seen these packets once, we don't
-             * care to see them again. If instead we forward them
-             * on the snoopfd (which points to the same device),
-             * then we will never get them back on snoopfd.
-             */
-	    if (sendto(snoopfd,pkt,clen,0,(struct sockaddr *)&to,sizeof(struct SOCKADDR)) < 0) {
-		syslog(LOG_ERR,"Error forwarding data packet to physical device from buffer: %m");
+
+	    wprot = *(unsigned short *)pkt;
+	    dpkt = pkt + sizeof(unsigned short);
+	    dlen = clen - sizeof(unsigned short);
+
+	    /* If we are using dynamic addresses we send the packet back
+	     * in to the kernel on the proxy so that it will go through
+	     * the firewall/masquerading rules again. It is up to the
+	     * user to enable and disable masquerading in ip-up/ip-down.
+	     * Masquerading a packet which was incorrectly masqueraded
+	     * to start with will never work :-).
+	     */
+	    if (dynamic_addrs) {
+		send_packet(wprot, dpkt, dlen);
+	    } else {
+#ifdef HAVE_AF_PACKET
+		if (af_packet)
+		    sl.sll_protocol = wprot;
+		else
+#endif
+		    sp.spkt_protocol = wprot;
+		if (sendto(snoopfd, dpkt, dlen, 0, to, to_len) < 0)
+		    mon_syslog(LOG_ERR,"Error forwarding data packet to physical device from buffer: %m");
 	    }
 	}
 	head = (head+6+clen)%buffer_size;
