@@ -2,15 +2,15 @@
  * diald.h - Main header file.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <stddef.h>
 #include <unistd.h>
-#include <strings.h>
+#include <string.h>
 #include <dirent.h>
 #include <fcntl.h>
 #include <ctype.h>
@@ -18,6 +18,9 @@
 #include <syslog.h>
 #include <signal.h>
 #include <time.h>
+#ifdef _POSIX_PRIORITY_SCHEDULING
+#  include <sched.h>
+#endif
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <sys/types.h>
@@ -52,28 +55,31 @@
 #include <arpa/inet.h>
 /* Hmm. Should there be a netinet pointer to these??? */
 #ifndef __GLIBC__ 
-#include <linux/if_ether.h>
-#include <linux/if_slip.h>
+#  include <linux/if_ether.h>
+#  include <linux/if_slip.h>
+#  include <linux/if_packet.h>
 #else
-#include <net/ethernet.h>
+#  include <net/ethernet.h>
+#  if __GLIBC_MINOR__ >= 1
+#    include <netpacket/packet.h>
+#  endif
+#  include <net/if_packet.h>
 #endif
 
-#include <linux/version.h>
-/* This only exists in kernels >= 1.3.75 */
-#if LINUX_VERSION_CODE >= 66379
-#define HAS_SOCKADDR_PKT
-#include <linux/if_packet.h>
-#define SOCKADDR sockaddr_pkt
+#if defined(AF_PACKET) && defined(PACKET_OUTGOING)
+#  define HAVE_AF_PACKET
+   typedef struct sockaddr_ll sockaddr_ll_t;
 #else
-#include <linux/socket.h>
-#define SOCKADDR sockaddr
+   typedef struct { int sll_protocol; } sockaddr_ll_t;
 #endif
 
 #include "config.h"
+#include "access.h"
 #include "fsm.h"
 #include "timer.h"
 #include "firewall.h"
 #include "bufio.h"
+#include "proxy.h"
 
 #define LOG_DDIAL	LOG_LOCAL2
 
@@ -104,42 +110,19 @@
 #define DEBUG_TICK		0x0020
 #define DEBUG_CONNECTION_QUEUE	0x0040
 
-/* Define MONITOR flags */
-#define MONITOR_STATE		0x0001
-#define MONITOR_INTERFACE	0x0002
-#define MONITOR_STATUS		0x0004
-#define MONITOR_LOAD		0x0008
-#define MONITOR_MESSAGE		0x0010
-#define MONITOR_QUEUE		0x0020
-
-/*
- * If you choose UNSAFE_ROUTING=0, then by default diald will route all
- * outgoing packets to the proxy device and forward them to the real
- * device by itself. This has the advantage that it gets around a bug
- * in the current production release (1.2.X) linux kernels
- * that causes TCP sessions to lock up if the route is changed while
- * a packet is being retransmitted. However, this introduces quite
- * a bit of overhead on outgoing packets (10-20%). (Note that incoming packets
- * don't go through this process!)
- * If you choose UNSAFE_ROUTING=1, then diald will change the routes anyway.
- * If you are using diald on a machine were there is a single outgoing
- * link this is perfectly safe (I think!), but if diald is being used
- * in an environment where more than one ppp or slip link can be active
- * at a time, then there is a small chance that you can lock up TCP
- * sessions. In particular if the link is terminated (either by diald
- * or by the other end hanging up) when a TCP session is in the middle
- * of retransmitting a packet, then that TCP session can become locked
- * if when the link comes back up it is brought back up on a different
- * ppp or slip device.
- * [NOTE: As of linux 1.3.13, the Linux kernel has been
- *  fixed to allow routing changes under an active TCP retransmit,
- *  so with 1.3.13 and later UNSAFE_ROUTING as the default is perfectly safe.]
- * [NOTE 2: unlike previous versions of diald, this option can now be
- * controlled from the command line or configuration file. See the new
- * options "reroute" and "-reroute".
+/* Define MONITOR flags
+ * 0x??????nn - status request flags, default on
+ * 0x??nnnn?? - status request flags, default off
+ * 0xnn?????? - syslog level
  */
-
-#define UNSAFE_ROUTING 1	/* do rerouting by default */
+#define MONITOR_STATE		0x00000001
+#define MONITOR_INTERFACE	0x00000002
+#define MONITOR_STATUS		0x00000004
+#define MONITOR_LOAD		0x00000008
+#define MONITOR_MESSAGE		0x00000010
+#define MONITOR_QUEUE		0x00000020
+#define MONITOR_VER1		0x00000080
+#define MONITOR_VER2		0x00000100
 
 /*
  * Originally diald just threw away any packets it received when
@@ -179,46 +162,72 @@
 #define DEFAULT_FIRST_PACKET_TIMEOUT 120
 #define DEFAULT_DIAL_DELAY 30
 #define DEFAULT_MTU 1500
+#define DEFAULT_METRIC 0
 #define DEFAULT_SPEED 38400
 
+#ifdef SCHED_OTHER
+#  define DEFAULT_SCHEDULER SCHED_OTHER
+#else
+#  define DEFAULT_SCHEDULER 0
+#endif
+#define DEFAULT_PRIORITY -5
+
 typedef struct monitors {
     struct monitors *next;
     int fd;			/* monitor output fp. */
-    int level;			/* Information level requested */
+    unsigned int level;		/* Information level requested */
     char *name;
 } MONITORS;
 
+
+#define SHELL_NOWAIT	0
+#define SHELL_WAIT	1
+extern int run_shell(int mode, const char *name, const char *buf, int fd);
+
 /* Configuration variables */
 
 char **devices;
 int device_count;
-char device[10];
-char device_node[9];
-int device_iface;
 int inspeed;
 int window;
 int mtu;
 int mru;
+int metric;
+char *link_name;
+char *link_desc;
+char *authsimple;
+char *initializer;
+char *deinitializer;
 char *connector;
 char *disconnector;
+char *proxyif;
 char *orig_local_ip;
 char *orig_remote_ip;
+char *orig_broadcast_ip;
 char *local_ip;
 unsigned long local_addr;
 char *remote_ip;
+char *broadcast_ip;
 char *netmask;
+char *ifsetup;
 char *addroute;
 char *delroute;
 char *ip_up;
 char *ip_down;
+char *ip_goingdown;
 char *acctlog;
 char *pidlog;
 char *fifoname;
+int tcpport;
+int demand;			/* enables demand dialling */
+int blocked;			/* user has blocked the link */
+int blocked_route;		/* blocked link has routes through it */
 char *lock_prefix;
 int pidstring;
 char *run_prefix;
 char *diald_config_file;
 char *diald_defs_file;
+char *path_ip;
 char *path_route;
 char *path_ifconfig;
 char *path_bootpc;
@@ -228,8 +237,9 @@ int buffer_size;
 int buffer_fifo_dispose;
 int buffer_timeout;
 FILE *acctfp;
-int call_start_time;
 int mode;
+int scheduler;
+int priority;
 int debug;
 int modem;
 int rotate_devices;
@@ -239,6 +249,7 @@ int strict_forwarding;
 int dynamic_addrs;
 int dynamic_mode;
 int slip_encap;
+int current_slip_encap;
 int lock_dev;
 int default_route;
 int pppd_argc;
@@ -260,8 +271,10 @@ int redial_backoff;
 int dial_fail_limit;
 int two_way;
 int give_way;
-int do_reroute;
 int proxyarp;
+#if 0
+int demasq;
+#endif
 int route_wait;
 
 #ifdef SIOCSKEEPALIVE
@@ -274,13 +287,13 @@ extern int outfill;
 
 /* Global variables */
 
+int af_packet;			/* kernel has AF_PACKET sockets */
 int fifo_fd;			/* FIFO command pipe. */
+int tcp_fd;			/* TCP listener. */
+fd_set ctrl_fds;		/* TCP command/monitor connections. */
+PIPE *pipes;			/* List of control/monitor pipes. */
 MONITORS *monitors;		/* List of monitor pipes. */
-int proxy_mfd;			/* master pty fd */
-FILE *proxy_mfp;		/* also have an fp. Hackery for recv_packet. */
-int proxy_sfd;			/* slave pty fd */
 int modem_fd;			/* modem device fp (for slip links) */
-char packet[4096];		/* slip packet buffer */
 int modem_hup;			/* have we seen a modem HUP? */
 int request_down;		/* has the user requested link down? */
 int request_up;			/* has the user requested link up? */
@@ -288,21 +301,24 @@ int forced;			/* has the user requested the link forced up? */
 int link_pid;			/* current pppd command pid */
 int dial_pid;			/* current dial command pid */
 int running_pid;		/* current system command pid */
+int running_status;		/* status of last system command */
 int dial_status;		/* status from last dial command */
 int state_timeout;		/* state machine timeout counter */
-int blocked;			/* user has blocked the link */
 int state;			/* DFA state */
 int current_retry_count;	/* current retry count */
-int proxy_iface;		/* Interface number for proxy pty */
+proxy_t proxy;			/* Proxy interface */
 int link_iface;			/* Interface number for ppp line */
 int orig_disc;			/* original PTY line disciple */
 int fwdfd;			/* control socket for packet forwarding */
 int snoopfd;			/* snooping socket fd */
 int fwunit;			/* firewall unit for firewall control */
 int req_pid;			/* pid of process that made "request" */
+char *current_dev;		/* name of the current device */
+int current_mode;		/* mode of the current link */
 char *req_dev;			/* name of the device file requested to open */
 int use_req;			/* are we actually using the FIFO link-up request device? */
 char snoop_dev[10];		/* The interface name we are listening on */
+int snoop_index;		/* The index of the interface */
 int txtotal,rxtotal;		/* transfer stats for the link */
 int itxtotal, irxtotal;		/* instantaneous transfer stats */
 int delayed_quit;		/* has the user requested delayed termination?*/
@@ -319,6 +335,7 @@ int ppp_half_dead;		/* is the ppp link half dead? */
 /* function prototypes */
 void init_vars(void);
 void parse_init(void);
+void parse_options_line(char *);
 void parse_options_file(char *);
 void parse_args(int, char *[]);
 void check_setup(void);
@@ -328,20 +345,22 @@ void block_signals(void);
 void unblock_signals(void);
 void filter_setup(void);
 void get_pty(int *, int *);
+int proxy_open(char *);
+void proxy_close(void);
 void proxy_up(void);
 void proxy_down(void);
-void proxy_config(char *, char *);
+void proxy_release(void);
 void dynamic_slip(void);
 void idle_filter_proxy(void);
 void open_fifo(void);
 void filter_read(void);
-void fifo_read(void);
+void ctrl_read(PIPE *);
 void proxy_read(void);
 void modem_read(void);
 void advance_filter_queue(void);
 void fire_timers(void);
 int recv_packet(unsigned char *, size_t);
-void send_packet(unsigned char *, size_t);
+void send_packet(unsigned short, unsigned char *, size_t);
 void sig_hup(int);
 void sig_intr(int);
 void sig_term(int);
@@ -357,17 +376,23 @@ void change_state(void);
 void output_state(void);
 void add_device(void *, char **);
 void set_str(char **, char **);
+void set_scheduler(char **, char **);
 void set_int(int *, char **);
 void set_flag(int *, char **);
+void set_flag2(int *, char **);
 void clear_flag(int *, char **);
+void set_blocked(int *, char **);
+void clear_blocked(int *, char **);
+void set_blocked_route(int *, char **);
+void clear_blocked_route(int *, char **);
 void set_mode(char **, char **);
 void set_dslip_mode(char **, char **);
 void read_config_file(int *, char **);
 void add_filter(void *var, char **);
 int insert_packet(unsigned char *, int);
-int lock(char *dev);
-void unlock(void);
-void fork_dialer(char *, int);
+char *lock(char *);
+void unlock(char *);
+void fork_dialer(char *, char *, int);
 void fork_connect(char *);
 void flush_timeout_queue(void);
 void set_up_tty(int, int, int);
@@ -386,6 +411,8 @@ void parse_up(void *var, char **argv);
 void parse_down(void *var, char **argv);
 void parse_prule(void *var, char **argv);
 void parse_var(void *var, char **argv);
+void iface_start(char *, char *, int, char *, char *, char *);
+void iface_stop(char *, char *, int, char *, char *, char *);
 void close_modem(void);
 int open_modem (void);
 void reopen_modem (void);
@@ -425,21 +452,18 @@ int fw_reset_wait(void);
 int next_alarm(void);
 void buffer_packet(unsigned int,unsigned char *);
 void forward_buffer(void);
-void run_ip_up(void);
-void run_ip_down(void);
-void set_ptp(char *, int, char *, int);
-void add_routes(char *, int, char *, char *, int);
-void del_routes(char *, int, char *, char *, int);
-void pipe_init(int, PIPE *);
+void run_state_script(char *name, char *script, int background);
+void pipe_init(char *, int, int, PIPE *, int);
 int pipe_read(PIPE *);
 void pipe_flush(PIPE *, int);
 int set_proxyarp (unsigned int);
 int clear_proxyarp (unsigned int);
 int report_system_result(int,char *);
-void mon_write(int,char *,int);
-void background_system(const char *);
+void mon_syslog(int pri, char *fmt, ...);
+void mon_write(unsigned int,char *,int);
 void block_timer();
 void unblock_timer();
+char *cdate(time_t);
 int getservice(const char *name, const char *proto);
 int getprotocol(const char *name);
 char *getprotonumber(int proto);
