@@ -2,12 +2,13 @@
  * firewall.c - Packet filtering for diald.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
 
 #include "diald.h"
-#ifdef __GLIBC__
+#if defined(__GLIBC__)
 typedef u_int8_t __u8;
 typedef u_int16_t __u16;
 typedef u_int32_t __u32;
@@ -38,7 +39,7 @@ static void init_units(void)
 	units[i].nrules = 0;
 	units[i].nfilters = 0;
 	if (!units[i].connections) {
-	    syslog(LOG_ERR,"Out of memory! AIIEEE!");
+	    mon_syslog(LOG_ERR,"Out of memory! AIIEEE!");
 	    die(1);
 	}
 	units[i].connections->next = units[i].connections->prev
@@ -54,15 +55,15 @@ static unsigned int in_slot(FW_Timeslot *slot, time_t *clock)
     struct tm *ltime = localtime(clock);
     int ctime = ltime->tm_hour*60*60+ltime->tm_min*60+ltime->tm_sec; 
 
-#ifdef 0
-    syslog(LOG_INFO,"slot check: %d %d %d %d",
+#if 0
+    mon_syslog(LOG_DEBUG,"slot check: %d %d %d %d",
 	ltime->tm_sec+ltime->tm_min*60+ltime->tm_hour*60*60, ltime->tm_wday,
 	ltime->tm_mday, ltime->tm_mon);
 #endif
 
     while (slot) {
-#ifdef 0
-    syslog(LOG_INFO,"slot def: %d %d %x %x %x",
+#if 0
+    mon_syslog(LOG_DEBUG,"slot def: %d %d %x %x %x",
 	slot->start, slot->end, slot->wday, slot->mday, slot->month);
 #endif 0
 	if ((slot->start <= ctime)
@@ -152,6 +153,144 @@ static unsigned int slot_end_timeout(FW_Timeslot *slot, time_t *clock)
     	return maxtime;
 }
 
+#if 0
+#ifdef __linux__
+/* Demasquerade an address. */
+static void
+demasquerade(int proto,
+	struct in_addr *saddr, int *sport,
+	struct in_addr *daddr, int *dport)
+{
+    struct in_addr *addr = NULL;
+    int *port = NULL;
+    FILE *fd;
+    int plen;
+    char protoname[8], pname[16], line[256];
+
+    /* If we are using an AF_PACKET socket then we know the source
+     * address is the first. If not all we know is that the first
+     * address is the numerically smaller.
+     *   Remember: only _one_ end of the link can be masqueraded
+     * by this system...
+     */
+
+    if (saddr->s_addr == local_addr
+#ifdef PORT_MASQ_BEGIN
+    && *sport >= PORT_MASQ_BEGIN
+#endif
+#ifdef PORT_MASQ_END
+    && *sport <= PORT_MASQ_END
+#endif
+    ) {
+	addr = saddr;
+	port = sport;
+    }
+
+    if (!addr
+#ifdef HAVE_AF_PACKET
+    && !af_packet
+#endif
+    && (daddr->s_addr == local_addr
+#ifdef PORT_MASQ_BEGIN
+    && *dport >= PORT_MASQ_BEGIN
+#endif
+#ifdef PORT_MASQ_END
+    && *dport <= PORT_MASQ_END
+#endif
+    )) {
+	addr = daddr;
+	port = dport;
+    }
+
+    if (!addr)
+	return;
+
+    if (!(fd = fopen("/proc/net/ip_masquerade", "r")))
+	return;
+
+    /* This matches what the kernel masquerade code says... */
+    switch (proto) {
+	case IPPROTO_UDP:
+	    strcpy(protoname, "UDP"); break;
+	case IPPROTO_TCP:
+	    strcpy(protoname, "TCP"); break;
+	case IPPROTO_ICMP:
+	    strcpy(protoname, "ICMP"); break;
+	default:
+	    sprintf(protoname, "IP_%d", proto); break;
+    }
+    plen = strlen(protoname);
+
+    sprintf(pname, "%04X", *port);
+
+    while (fgets(line, sizeof(line), fd)) {
+	if (strncmp(line, protoname, plen))
+	    continue;
+
+	if (!strncmp(line+plen+29, pname, 4)) {
+	    addr->s_addr = htonl(strtoul(line+plen+1, NULL, 16));
+	    *port = strtoul(line+plen+10, NULL, 16);
+	    break;
+	}
+    }
+    fclose(fd);
+}
+#endif
+#endif
+
+/* Generate a connection description */
+static char *
+desc_connection(FW_Connection *c)
+{
+    struct in_addr saddr, daddr;
+    int sport, dport;
+    char sad_text[20], dad_text[20];
+    char *pent;
+    char proto[20];
+    char *buf;
+
+    /* The packet has network byte order, we want to preserve it
+     * in the address regardless of the host ordering.
+     */
+    saddr.s_addr = htonl((c->id.id[1] << 24)
+			+ (c->id.id[2] << 16)
+			+ (c->id.id[3] << 8)
+			+ (c->id.id[4]));
+    sport = c->id.id[10]+(c->id.id[9]<<8);
+    daddr.s_addr = htonl((c->id.id[5] << 24)
+			+ (c->id.id[6] << 16)
+			+ (c->id.id[7] << 8)
+			+ (c->id.id[8]));
+    dport = c->id.id[12]+(c->id.id[11]<<8);
+
+#if 0
+#if defined(__linux__)
+    if (demasq)
+	demasquerade(c->id.id[0], &saddr, &sport, &daddr, &dport);
+#endif
+#endif
+
+    strcpy(sad_text, inet_ntoa(saddr));
+    strcpy(dad_text, inet_ntoa(daddr));
+
+    pent = getprotonumber(c->id.id[0]);
+    if (pent) {
+	strncpy(proto, pent, sizeof(proto)-1);
+	proto[sizeof(proto)-1] = '\0';
+    } else {
+	sprintf(proto, "%d", c->id.id[0]);
+    }
+
+    buf = malloc(64);
+    if (buf) {
+	snprintf(buf, 64-1, "%-4s  %15s/%-5d  %15s/%-5d",
+		proto, sad_text, sport, dad_text, dport);
+	buf[64-1] = '\0';
+    }
+
+    return buf;
+}
+
 /* Find a connection in the queue */
 static FW_Connection *find_connection(FW_unit *unit, FW_ID *id)
 {
@@ -175,8 +314,9 @@ static FW_Connection *find_connection(FW_unit *unit, FW_ID *id)
  * Add/update a connection in the queue.
  */
 
-static void add_connection(
-    FW_unit *unit, FW_Connection *c, FW_ID *id, unsigned int timeout, TCP_STATE lflags)
+static void add_connection(FW_unit *unit, FW_Connection *c, FW_ID *id,
+			unsigned int timeout, TCP_STATE lflags,
+			int direction, int len)
 {
     /* look for a connection that matches this one */
     if (c == NULL) {
@@ -184,15 +324,27 @@ static void add_connection(
 	    /* no matching connection, add one */
 	    c = malloc(sizeof(FW_Connection));
 	    if (c == 0) {
-	       syslog(LOG_ERR,"Out of memory! AIIEEE!");
+	       mon_syslog(LOG_ERR,"Out of memory! AIIEEE!");
 	       die(1);
 	    }
 	    c->id = *id;
+	    c->description = NULL;
+	    c->packets[0] = c->packets[1] = 0;
+	    c->bytes[0] = c->bytes[1] = 0;
+	    c->bytes_total[0] = c->bytes_total[1] = 0;
+	    c->packets[direction] = 1;
+	    c->bytes[direction] = len;
 	    init_timer(&c->timer);
             c->tcp_state = lflags;
 	    c->unit = unit;
 	    c->timer.data = (void *)c;
 	    c->timer.function = (void *)(void *)del_connection;
+	    if (unit->connections->next == unit->connections
+	    && state != STATE_UP && !blocked && demand) {
+		c->description = desc_connection(c);
+		if (c->description)
+		    mon_syslog(LOG_NOTICE, "Trigger: %s", c->description);
+	    }
 	    c->next = unit->connections->next;
 	    c->prev = unit->connections;
 	    unit->connections->next->prev = c;
@@ -200,7 +352,7 @@ static void add_connection(
 	    c->timer.expires = timeout;
 	    add_timer(&c->timer);
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-    		syslog(LOG_INFO,"Adding connection %p @ %ld - timeout %d",c,
+    		mon_syslog(LOG_DEBUG,"Adding connection %p @ %ld - timeout %d",c,
 			time(0),timeout);
 	}
     } else {
@@ -208,10 +360,12 @@ static void add_connection(
 	if (timeout > 0) {
 	    /* reanimating a ghost? */
 	    del_timer(&c->timer);
+	    c->packets[direction]++;
+	    c->bytes[direction] += len;
 	    c->timer.expires = timeout;
 	    add_timer(&c->timer);
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-    		syslog(LOG_INFO,"Adding connection %p @ %ld - timeout %d",c,
+    		mon_syslog(LOG_DEBUG,"Adding connection %p @ %ld - timeout %d",c,
 			time(0),timeout);
 	} else {
 	    /* timeout = 0, so toss the connection */
@@ -228,10 +382,11 @@ static void add_connection(
 void del_connection(FW_Connection *c)
 {
     if (debug&DEBUG_CONNECTION_QUEUE)
-	syslog(LOG_INFO,"Deleting connection %p @ %ld",c,time(0));
+	mon_syslog(LOG_DEBUG,"Deleting connection %p @ %ld",c,time(0));
 
     c->next->prev = c->prev;
     c->prev->next = c->next;
+    if (c->description) free(c->description);
     free(c);
 }
 
@@ -245,7 +400,7 @@ void del_impulse(FW_unit *unit)
 	    unit->impulse.function = (void *)(void *)del_impulse;
 	    unit->impulse.expires = impulse_time;
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-		syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
+		mon_syslog(LOG_DEBUG,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
 	    add_timer(&unit->impulse);
 	}
     } else {
@@ -256,7 +411,7 @@ void del_impulse(FW_unit *unit)
 	    unit->impulse.function = (void *)(void *)del_impulse;
 	    unit->impulse.expires = impulse_fuzz;
 	    if (debug&DEBUG_CONNECTION_QUEUE)
-		syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
+		mon_syslog(LOG_DEBUG,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
 	    add_timer(&unit->impulse);
 	}
     }
@@ -373,7 +528,7 @@ next_rule: /* try the next filter */
     	unit->impulse.expires = (force)?ftimeout:itimeout;
     	add_timer(&unit->impulse);
 	if (debug&DEBUG_CONNECTION_QUEUE)
-	    syslog(LOG_INFO,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
+	    mon_syslog(LOG_DEBUG,"Refreshing impulse generator: mode %d, time %ld @ %ld",unit->impulse_mode,unit->impulse.expires,time(0));
     } else {
 	/* set these so that the monitor output is sane */
 	unit->impulse_mode = 1;
@@ -398,10 +553,10 @@ static void log_packet(int accept, struct iphdr *pkt, int len,  int rule)
 	sport = ntohs(udp->source), dport = ntohs(udp->dest);
 
     if (pkt->protocol == IPPROTO_TCP) {
-	syslog(LOG_INFO,
-	    "filter %s rule %d proto %d len %d seq %lx ack %lx flags %s%s%s%s%s%s packet %s,%d => %s,%d",
-	    (accept)?"accepted":"ignored",rule,
-	    pkt->protocol,
+	mon_syslog(LOG_DEBUG,
+	    "filter %s rule %d proto tcp len %d seq %lx ack %lx flags %s%s%s%s%s%s packet %s,%d => %s,%d",
+	    (accept)?"accepted":"ignored",
+	    rule,
 	    ntohs(pkt->tot_len),
 	    ntohl(tcp->seq), ntohl(tcp->ack_seq),
 	    (tcp->fin) ? " FIN" : "",
@@ -412,10 +567,11 @@ static void log_packet(int accept, struct iphdr *pkt, int len,  int rule)
 	    (tcp->urg) ? " URG" : "",
 	    saddr, sport, daddr, dport);
     } else {
-	syslog(LOG_INFO,
+	mon_syslog(LOG_DEBUG,
 	    "filter %s rule %d proto %d len %d packet %s,%d => %s,%d",
-	    (accept)?"accepted":"ignored",rule,
-	    pkt->protocol,
+	    (accept)?"accepted":"ignored",
+	    rule,
+	    getprotonumber(pkt->protocol),
 	    htons(pkt->tot_len),
 	    saddr, sport, daddr, dport);
     }
@@ -452,11 +608,11 @@ static void ip_swap_addrs(struct iphdr *pkt)
 void print_filter(FW_Filter *filter)
 {
     int i;
-    syslog(LOG_INFO,"filter: prl %d log %d type %d cnt %d tm %d",
+    mon_syslog(LOG_DEBUG,"filter: prl %d log %d type %d cnt %d tm %d",
 	filter->prule,filter->log,filter->type,
 	filter->count,filter->timeout);
     for (i = 0; i < filter->count; i++) {
-	syslog(LOG_INFO,"    term: shift %d op %d off %d%c msk %x tst %x",
+	mon_syslog(LOG_DEBUG,"    term: shift %d op %d off %d%c msk %x tst %x",
 	    filter->terms[i].shift, filter->terms[i].op,
 	    filter->terms[i].offset&0x7f,
 	    (filter->terms[i].offset&0x80)?'d':'h',
@@ -477,7 +633,7 @@ __u16 checksum(__u16 *buf, int nwords)
 }
 
 
-void forge_tcp_reset(struct iphdr *iph, int len)
+void forge_tcp_reset(unsigned short wprot, struct iphdr *iph, int len)
 {
     struct tcphdr *tcph,*ntcph;
     struct iphdr *niph;
@@ -553,13 +709,14 @@ void forge_tcp_reset(struct iphdr *iph, int len)
     /*
      * OK, send the packet now.
      */
-    send_packet((unsigned char *)niph,sizeof(struct tcphdr)+sizeof(struct iphdr));
+    send_packet(wprot, (unsigned char *)niph,
+	sizeof(struct tcphdr)+sizeof(struct iphdr));
 
     free((void *)niph);
 }
 
 /* Check if a packet passes the filters */
-int check_firewall(int unitnum, unsigned char *pkt, int len)
+int check_firewall(int unitnum, sockaddr_ll_t *sll, unsigned char *pkt, int len)
 {
     FW_unit *unit;
     FW_Filters *fw;
@@ -603,19 +760,32 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
      /* If we have the "strict-forwarding" option set then unless one of
      * the two addresses matches the local address we should ignore it.
      */
-    if (strict_forwarding && ip_pkt->saddr != local_addr && ip_pkt->daddr != local_addr) {
+    if (strict_forwarding
+    && (!sll || sll->sll_protocol == htons(ETH_P_IP))
+    && ip_pkt->saddr != local_addr && ip_pkt->daddr != local_addr) {
 	/* We forge resets for TCP protocol stuff that has been
 	 * shut down do to a link failure.
 	 */
 	if (ip_pkt->protocol == IPPROTO_TCP)
-	    forge_tcp_reset(ip_pkt,len);
+	    forge_tcp_reset(htons(ETH_P_IP), ip_pkt, len);
 	
 	goto skip;
     }
 
     /* Build the connection ID, and set the direction flag */
-    direction = ip_direction(ip_pkt);
-    if (direction == 2) ip_swap_addrs(ip_pkt);
+#ifdef HAVE_AF_PACKET
+    if (af_packet) {
+	direction = 1;
+	if (sll->sll_pkttype != PACKET_OUTGOING) {
+	    direction = 2;
+	    ip_swap_addrs(ip_pkt);
+	}
+    } else
+#endif
+    {
+	direction = ip_direction(ip_pkt);
+	if (direction == 2) ip_swap_addrs(ip_pkt);
+    }
 
     memset(&id,0,sizeof(id));
     for (i = 0; i < FW_ID_LEN; i++)
@@ -629,7 +799,11 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
     /* Do the TCP liveness changes */
     if (ip_pkt->protocol == IPPROTO_TCP) {
 	struct tcphdr *tcp = (struct tcphdr *)((char *)ip_pkt + 4*ip_pkt->ihl);
+#if 1
+	int tcp_data_len = len - (4*ip_pkt->ihl + tcp->doff*4);
+#else
 	int tcp_data_len = len - (4*ip_pkt->ihl + sizeof(struct tcphdr));
+#endif
 
 	if (conn) {
 	    lflags = conn->tcp_state;
@@ -666,7 +840,7 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 
     rule = 1;
     while (fw) {
-#ifdef 0
+#if 0
 	print_filter(&fw->filt);
 #endif
 	/* is this rule currently applicable? */
@@ -694,12 +868,15 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 	    term = &fw->filt.terms[i];
 	    if (FW_TCP_STATE(term->offset))
 		v = (lflags.tcp_flags >> term->shift) && term->mask;
-	    else
-	        v = (ntohl(*(int *)(&(FW_IN_DATA(term->offset)?data:pkt)
-				  [FW_OFFSET(term->offset)]))
-		    >> term->shift) & term->mask;
-#ifdef 0
-	    syslog(LOG_INFO,"testing ip %x:%x data %x:%x mask %x shift %x test %x v %x",
+	    else {
+		int n;
+	        memcpy(&n, &(FW_IN_DATA(term->offset)?data:pkt)
+				  [FW_OFFSET(term->offset)],
+			sizeof(int));
+	        v = (ntohl(n) >> term->shift) & term->mask;
+	    }
+#if 0
+	    mon_syslog(LOG_DEBUG,"testing ip %x:%x data %x:%x mask %x shift %x test %x v %x",
 		ntohl(*(int *)(&pkt[FW_OFFSET(term->offset)])),
 		*(int *)(&pkt[FW_OFFSET(term->offset)]),
 		ntohl(*(int *)(&data[FW_OFFSET(term->offset)])),
@@ -717,7 +894,7 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 	    }
 	}
 	/* Ok, we matched a rule. What are we suppose to do? */
-#ifdef 0
+#if 0
 	if (fw->filt.log)
 #endif
         if (debug&DEBUG_FILTER_MATCH)
@@ -730,7 +907,9 @@ int check_firewall(int unitnum, unsigned char *pkt, int len)
 		conn,
 		&id,
 		fw->filt.timeout,
-		lflags);
+		lflags,
+		direction-1,
+		len);
 	}
 	/* check if we are no longer waiting */
 	if (fw->filt.type == FW_TYPE_WAIT) {
@@ -756,14 +935,13 @@ skip:
     return 1;
 }
 
-static void pcountdown(int level, long secs)
+static char * pcountdown(char *buf, long secs)
 {
-    char buf[10];
     /* Make sure that we don't try to print values that overflow our buffer */
     if (secs < 0) secs = 0;
     if (secs > 359999) secs = 359999;
-    sprintf(buf,"%02ld:%02ld:%02ld\n",secs/3600,(secs/60)%60,secs%60);
-    if (monitors) mon_write(level,buf,9);
+    sprintf(buf,"%02ld:%02ld:%02ld",secs/3600,(secs/60)%60,secs%60);
+    return buf;
 }
 
 int ctl_firewall(int op, struct firewall_req *req)
@@ -787,9 +965,7 @@ int ctl_firewall(int op, struct firewall_req *req)
 	    c != unit->connections; c = cn) {
 		cn = c->next;
 		del_timer(&c->timer);
-		c->next->prev = c->prev;
-		c->prev->next = c->next;
-		free((void *)c);
+		del_connection(c);
 	    }
 	    return 0;
 	}
@@ -826,7 +1002,7 @@ int ctl_firewall(int op, struct firewall_req *req)
 	{
 	    FW_Filters *filters = malloc(sizeof(FW_Filters));
 	    if (filters == 0) {
-		syslog(LOG_ERR,"Out of memory! AIIEEE!");
+		mon_syslog(LOG_ERR,"Out of memory! AIIEEE!");
 		return -1; /* ERRNO */
 	    }
 	    filters->next = 0;
@@ -851,7 +1027,7 @@ int ctl_firewall(int op, struct firewall_req *req)
 	    FW_Connection *c;
 	    char saddr[20], daddr[20];
     	    struct in_addr addr;
-	    syslog(LOG_INFO,"up = %d, forcing = %d, impulse = %d, iitime = %d, itime = %d, ifuzz = %d, itimeout = %ld, timeout = %ld, next alarm = %d",
+	    mon_syslog(LOG_DEBUG,"up = %d, forcing = %d, impulse = %d, iitime = %d, itime = %d, ifuzz = %d, itimeout = %ld, timeout = %ld, next alarm = %d",
 		unit->up,unit->force, unit->impulse_mode, impulse_init_time, impulse_time,
 		impulse_fuzz,
 		unit->impulse.expected-tstamp,unit->force_etime-atime,
@@ -864,7 +1040,7 @@ int ctl_firewall(int op, struct firewall_req *req)
                 addr.s_addr = c->id.id[5] + (c->id.id[6]<<8)
                         + (c->id.id[7]<<16) + (c->id.id[8]<<24);
                 strcpy(daddr,inet_ntoa(addr));
-                syslog(LOG_INFO,
+                mon_syslog(LOG_DEBUG,
                         "ttl %ld, %d - %s/%d => %s/%d (tcp state ([%lx,%lx] %d,%d))",
                         c->timer.expected-tstamp, c->id.id[0],
                         saddr, c->id.id[10]+(c->id.id[9]<<8),
@@ -877,55 +1053,70 @@ int ctl_firewall(int op, struct firewall_req *req)
 	    return 0;
 	}
 	return 0;
+    case IP_FW_MCONN_INIT:
+	if (!req) return -1;
+	{
+	    FW_Connection *c;
+	    for (c=unit->connections->next; c!=unit->connections; c=c->next) {
+		c->bytes_total[0] += c->bytes[0];
+		c->bytes_total[1] += c->bytes[1];
+		c->packets[0] = c->packets[1] = c->bytes[0] = c->bytes[1] = 0;
+	    }
+	}
+	return 0;
     case IP_FW_MCONN:
 	if (!req || !monitors) return -1; /* ERRNO */
 	{
 	    unsigned long atime = time(0);
             unsigned long tstamp = timestamp();
 	    FW_Connection *c;
-	    char saddr[20], daddr[20];
-  	    int sport, dport;
-	    char proto[20];
-    	    char *pent;
-    	    struct in_addr addr;
+	    int i;
+	    char *p;
+	    char tbuf1[10], tbuf2[10], tbuf3[10];
             char buf[1024];
 
-	    sprintf(buf,"STATUS\n%d\n%d\n%d\n%d\n%d\n%d\n",
+	    sprintf(buf,"STATUS\n%d %d %d %d %d %d %s %s %s %c %c %c\n",
 		unit->up, unit->force, unit->impulse_mode,
 		impulse_init_time, impulse_time,
-		impulse_fuzz);
-	    if (monitors) mon_write(MONITOR_STATUS,buf,strlen(buf));
-	    /* If this is non-zero we just have an expired timer so
-	     * we of course display nothing.
-	     */
-            if (unit->impulse.expected-tstamp > 0)
-                pcountdown(MONITOR_STATUS,unit->impulse.expected-tstamp);
-            else
-                pcountdown(MONITOR_STATUS,0);
-	    pcountdown(MONITOR_STATUS,unit->force_etime-atime);
-	    pcountdown(MONITOR_STATUS,next_alarm());
-	    if (monitors) mon_write(MONITOR_QUEUE,"QUEUE\n",6);
+		impulse_fuzz,
+		((unit->impulse.expected-tstamp > 0)
+		    ? pcountdown(tbuf1, unit->impulse.expected-tstamp)
+		    : pcountdown(tbuf1, 0)),
+		pcountdown(tbuf2, unit->force_etime-atime),
+		pcountdown(tbuf3, next_alarm()),
+		(demand ? '1' : '0'),
+		(blocked ? '1' : '0'),
+		(forced ? '1' : '0')
+	    );
+	    mon_write(MONITOR_VER2|MONITOR_STATUS, buf, strlen(buf));
+
+	    p = buf + 7;
+	    for (i=9; i > 0; i--) {
+		while (*p != ' ') p++;
+		*p = '\n';
+	    }
+	    *(p+1) = '\0';
+	    mon_write(MONITOR_VER1|MONITOR_STATUS, buf, strlen(buf));
+
+	    mon_write(MONITOR_QUEUE,"QUEUE\n",6);
 	    for (c=unit->connections->next; c!=unit->connections; c=c->next) {
 		if (c->timer.next == 0) continue;
-                addr.s_addr = c->id.id[1] + (c->id.id[2]<<8)
-                        + (c->id.id[3]<<16) + (c->id.id[4]<<24);
-                strcpy(saddr,inet_ntoa(addr));
-                addr.s_addr = c->id.id[5] + (c->id.id[6]<<8)
-                        + (c->id.id[7]<<16) + (c->id.id[8]<<24);
-                strcpy(daddr,inet_ntoa(addr));
-		pent = getprotonumber(c->id.id[0]);
-		if (pent) strcpy(proto,pent);
-		else sprintf(proto,"%d",c->id.id[0]);
-		sport = c->id.id[10]+(c->id.id[9]<<8);
-		dport = c->id.id[12]+(c->id.id[11]<<8);
-                sprintf(buf,
-                        "%-4s  %15s/%-5d  %15s/%-5d  ",
-                        proto, saddr, sport,
-                        daddr, dport);
-                if (monitors) mon_write(MONITOR_QUEUE,buf,strlen(buf));
-		pcountdown(MONITOR_QUEUE,c->timer.expected-tstamp);
+		if (!c->description) c->description = desc_connection(c);
+		strcpy(buf, c->description);
+                sprintf(buf+50,
+                        "  %8.8s %ld %ld %ld %ld %ld %ld\n",
+			pcountdown(tbuf1, c->timer.expected-tstamp),
+			c->packets[0], c->bytes[0], c->bytes_total[0],
+			c->packets[1], c->bytes[1], c->bytes_total[1]);
+		c->bytes_total[0] += c->bytes[0];
+		c->bytes_total[1] += c->bytes[1];
+		c->packets[0] = c->packets[1] = c->bytes[0] = c->bytes[1] = 0;
+                mon_write(MONITOR_VER2|MONITOR_QUEUE, buf, strlen(buf));
+		buf[60] = '\n';
+		buf[61] = '\0';
+                mon_write(MONITOR_VER1|MONITOR_QUEUE, buf, strlen(buf));
 	    }
-	    if (monitors) mon_write(MONITOR_QUEUE,"END QUEUE\n",10);
+	    mon_write(MONITOR_QUEUE,"END QUEUE\n",10);
 	    return 0;
 	}
 	return 0;
@@ -966,13 +1157,15 @@ int ctl_firewall(int op, struct firewall_req *req)
 	}
     case IP_FW_UP:
 	unit->up = 1;
-	run_ip_up();
+	if (ip_up)
+	    run_state_script("ip-up", ip_up, 1);
 	fw_force_update(unit);
 	fw_impulse_update(unit,1);
 	return 0;
 
     case IP_FW_DOWN:
-	if (unit->up) run_ip_down();
+	if (unit->up && ip_down)
+	    run_state_script("ip-down", ip_down, 1);
 	unit->up = 0;
 	/* turn off the impulse generator */
 	del_timer(&unit->impulse);
