@@ -2,12 +2,16 @@
  * fsm.c - Control Finite State Machine for diald main loop.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
 
 #include "diald.h"
 
+int call_timer_running = 0;
+time_t call_start_time;
+
 #define STATE(name,timeout,tstate) \
         { timeout, act_ ## name, trans_ ## name, STATE_ ## tstate, #name }
 
@@ -32,8 +36,8 @@ struct {
 
 void validate_mode()
 {
-	if (mode < 0 || mode > 2) {
-		syslog(LOG_ERR, "Mode value strayed to %d. Tell Eric.\n",mode);
+	if (current_mode < 0 || current_mode > 2) {
+		mon_syslog(LOG_ERR, "Mode value strayed to %d. Tell Eric.\n",current_mode);
 		die(1);
 	}
 	if (pcontrol[0].start != slip_start
@@ -57,20 +61,19 @@ void validate_mode()
 	|| pcontrol[2].reroute != dev_reroute
 	|| pcontrol[2].kill != dev_kill
 	|| pcontrol[2].zombie != dev_zombie) {
-		syslog(LOG_ERR, "Corrupt FSM table. Tell Eric.\n");
+		mon_syslog(LOG_ERR, "Corrupt FSM table. Tell Eric.\n");
 		die(1);
 	}
 }
 
-void control_start() { validate_mode(); (*pcontrol[mode].start)(); }
-int control_set_addrs() { validate_mode(); return (*pcontrol[mode].set_addrs)(); }
-int control_dead() { validate_mode(); return (*pcontrol[mode].dead)(); }
-void control_stop() { validate_mode(); (*pcontrol[mode].stop)(); }
-void control_reroute() { validate_mode(); (*pcontrol[mode].reroute)(); }
-void control_kill() { validate_mode(); (*pcontrol[mode].kill)(); }
-void control_zombie() { validate_mode(); (*pcontrol[mode].zombie)(); }
+void control_start() { validate_mode(); (*pcontrol[current_mode].start)(); }
+int control_set_addrs() { validate_mode(); return (*pcontrol[current_mode].set_addrs)(); }
+int control_dead() { validate_mode(); return (*pcontrol[current_mode].dead)(); }
+void control_stop() { validate_mode(); (*pcontrol[current_mode].stop)(); }
+void control_reroute() { validate_mode(); (*pcontrol[current_mode].reroute)(); }
+void control_kill() { validate_mode(); (*pcontrol[current_mode].kill)(); }
+void control_zombie() { validate_mode(); (*pcontrol[current_mode].zombie)(); }
 
-int blocked = 0;		/* user has blocked the link */
 int state = STATE_DOWN;		/* DFA state */
 int current_retry_count = 0;	/* current retry count */
 int no_redial_delay = 0;
@@ -89,37 +92,30 @@ void trans_DOWN(void)
 {
     request_down = 0;
     use_req = 0;
-    if (request_up && (req_pid || !blocked)) {
-	request_up = 0; GOTO(STATE_CONNECT);
+    if (request_up) {
+	request_up = 0;
+        if (req_pid || !blocked)
+	    GOTO(STATE_CONNECT);
+	else
+	    mon_syslog(LOG_NOTICE, "Blocked. Up request ignored");
     }
-    else if ((forced || !queue_empty()) && !blocked) GOTO(STATE_CONNECT);
+    else if ((forced || (!queue_empty() && demand)) && !blocked)
+	GOTO(STATE_CONNECT);
     else if (delayed_quit && queue_empty() && !forced) {
-	syslog(LOG_INFO,"Carrying out delayed termination request.");
+	mon_syslog(LOG_NOTICE,"Carrying out delayed termination request.");
 	terminate = 1;
     }
 }
 
-/* Grumble. Attempt to generate a nicely formatted ascii date without
- * a built in newline.
- */
-char *cdate(void)
-{
-    static char dt[128];
-    int len;
-
-    time((time_t *)&call_start_time);
-    len = strftime(dt,128,"%c %Z",localtime((time_t *)&call_start_time));
-    if (len == 128) dt[len] = 0;
-    return dt;
-}
-
 void act_CONNECT(void)
 {
-    if (!req_pid && acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
-
-	    fprintf(acctfp,"%s: Calling site %s.\n",
-	    cdate(), remote_ip);
+    if (!req_pid) {
+	mon_syslog(LOG_NOTICE, "Calling site %s\n", remote_ip);
+	if (acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
+	    fprintf(acctfp, "%s: Calling site %s.\n",
+		cdate(time(0)), remote_ip);
 	    fclose(acctfp);
+	}
     }
     txtotal = rxtotal = 0;
     dial_status = 0;
@@ -134,11 +130,10 @@ void trans_CONNECT(void)
 	if (dial_status == 0) {
 	    dial_failures = 0;
 	    redial_rtimeout = redial_timeout;
-	    if (mode != MODE_DEV)
-	        finish_dial(); /* go into a CLOCAL connection */
+	    finish_dial(); /* go into a CLOCAL connection */
 	    GOTO(STATE_START_LINK);
 	} else {
-	    syslog(LOG_INFO,"Connect script failed.");
+	    mon_syslog(LOG_NOTICE,"Connect script failed.");
 	    dial_failures++;
 	    if (redial_rtimeout == -1) redial_rtimeout = redial_timeout;
 	    if (redial_backoff_start >= 0 && dial_failures > redial_backoff_start) {
@@ -148,19 +143,24 @@ void trans_CONNECT(void)
 	    }
 	    if (dial_fail_limit != 0 && dial_failures > dial_fail_limit) {
 		/* want to block connection until the user comes back...*/
+		/* FIXME: should we just disable demand dialling? */
 		blocked = 1;
 		dial_failures = 0;
 		redial_rtimeout = redial_timeout;
 		current_retry_count = 0;
-		syslog(LOG_INFO,"Too many dialing failures in a row. Blocking connection.");
+		mon_syslog(LOG_NOTICE,"Too many dialing failures in a row. Blocking connection.");
 	    }
 	    GOTO(STATE_CLOSE);
 	}
+    } else if (request_down) {
+        mon_syslog(LOG_NOTICE,"Cancelling connect script.");
+	GOTO(STATE_STOP_DIAL);
     } else if (state_timeout == 0) {
-	syslog(LOG_INFO,"Connect script timed out. Killing script.");
+	mon_syslog(LOG_NOTICE,"Connect script timed out. Killing script.");
+	GOTO(STATE_STOP_DIAL);
     } else if (req_pid && !use_req) {
         /* we never get here on a connect FIFO request anyway */
-        syslog(LOG_INFO,"Cancelling connect script in favour of FIFO request.");
+        mon_syslog(LOG_NOTICE,"Cancelling connect script in favour of request.");
 	GOTO(STATE_STOP_DIAL);
     }
 }
@@ -168,10 +168,14 @@ void trans_CONNECT(void)
 void act_STOP_DIAL(void)
 {
     if (dial_pid) {
+#if 0
         if (debug&DEBUG_VERBOSE)
-            syslog(LOG_INFO,"Sending SIGINT to (dis)connect process %d",
+#endif
+            mon_syslog(LOG_DEBUG,"Sending SIGINT to (dis)connect process %d",
 		dial_pid);
-	if (kill(dial_pid,SIGINT) == -1 && errno == ESRCH) {
+	if (kill(-dial_pid, SIGINT) == -1
+	&& kill(dial_pid, SIGINT) == -1
+	&& errno == ESRCH) {
 	    dial_pid = 0;
 	    dial_status = -1;
 	}
@@ -186,10 +190,14 @@ void trans_STOP_DIAL(void)
 void act_KILL_DIAL(void)
 {
     if (dial_pid) {
+#if 0
         if (debug&DEBUG_VERBOSE)
-            syslog(LOG_INFO,"Sending SIGKILL to (dis)connect process %d",
+#endif
+            mon_syslog(LOG_DEBUG,"Sending SIGKILL to (dis)connect process %d",
 		dial_pid);
-	if (kill(dial_pid,SIGKILL) == -1 && errno == ESRCH) {
+	if (kill(-dial_pid, SIGKILL) == -1
+	&& kill(dial_pid, SIGKILL) == -1
+	&& errno == ESRCH) {
 	    dial_pid = 0;
 	    dial_status = -1;
 	}
@@ -202,13 +210,20 @@ void trans_KILL_DIAL(void)
 
 void act_START_LINK(void)
 {
+    time(&call_start_time);
+    call_timer_running = 1;
+    mon_syslog(LOG_NOTICE,
+	use_req ? "Connect request from %s\n" : "Connected to site %s\n",
+	remote_ip);
+
     if (acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
         fprintf(acctfp,"%s: %s %s.\n",
-	    cdate(),
+	    cdate(call_start_time),
 	    use_req ? "Connect request from" : "Connected to site",
 	    remote_ip);
 	fclose(acctfp);
     }
+
     validate_mode();
     control_start();
 }
@@ -229,7 +244,7 @@ void trans_START_LINK(void)
 	/* If we time out, then we're going to stop. We need to reroute
          *  just in case the routes got changed before we timed out.
 	 */
-	syslog(LOG_INFO,"pppd startup timed out. Check your pppd options. Killing pppd.");
+	mon_syslog(LOG_NOTICE,"pppd startup timed out. Check your pppd options. Killing pppd.");
 	control_reroute();
     }
 }
@@ -260,10 +275,14 @@ void act_UP(void)
     ppp_half_dead = 0;
     if (buffer_packets)
    	 forward_buffer();
-    if (use_req && acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
-        fprintf(acctfp,"%s: Connected to site %s.\n",
-	    cdate(), remote_ip);
-	fclose(acctfp);
+    if (use_req) {
+	mon_syslog(LOG_NOTICE, "Connected to site %s\n", remote_ip);
+
+	if (acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
+	    fprintf(acctfp, "%s: Connected to site %s.\n",
+		cdate(time(0)), remote_ip);
+	    fclose(acctfp);
+	}
     }
     /* Once we get here, we might as well kill off any outstanding
      * FIFO requests */
@@ -271,20 +290,20 @@ void act_UP(void)
         killpg(req_pid, SIGKILL);
         kill(req_pid, SIGKILL);
 	req_pid=0;
-        syslog(LOG_INFO,"Cancelling link up requested denied in favour of existing link.");
+        mon_syslog(LOG_NOTICE,"Cancelling link up request in favour of existing link.");
     }
 }
 
 void trans_UP(void)
 {
     request_up = 0;
-    if (blocked || request_down) {
+    if (request_down) {
 	request_down = 0;
 	goto take_link_down;
     }
     if (state_timeout == 0) {
 	if (!forced && queue_empty()) {  /* the link may be forced up. */
-	    syslog(LOG_INFO,"%s %d %s",
+	    mon_syslog(LOG_NOTICE,"%s %d %s",
 	    	"Failed to received first packet within",
 	    	first_packet_timeout,
 	    	"seconds. Closing Link down.");
@@ -293,8 +312,10 @@ void trans_UP(void)
 	state_timeout = -1;
     }
     if (!forced && queue_empty() && state_timeout == -1) {
-	syslog(LOG_INFO,"Closing down idle link.");
+	mon_syslog(LOG_NOTICE,"Closing down idle link.");
 take_link_down:
+	if (ip_goingdown)
+	    run_state_script("ip-goingdown", ip_goingdown, 0);
 	current_retry_count = 0;
 	flush_timeout_queue();
 	/* WARNING: You MUST do this before calling idle_filter_proxy(),
@@ -305,12 +326,12 @@ take_link_down:
 	return;
     }
     if (ppp_half_dead) {
-	syslog(LOG_ERR,"PPP network layer died, but link did not. Probable configuration error.");
+	mon_syslog(LOG_ERR,"PPP network layer died, but link did not. Probable configuration error.");
 	GOTO(STATE_HALF_DEAD);
 	return;
     }
     if (control_dead() || modem_hup) {
-	syslog(LOG_INFO,"Link died on remote end.");
+	mon_syslog(LOG_NOTICE,"Link died on remote end.");
     	if (two_way) flush_timeout_queue();
 	no_redial_delay = 1;
 	current_retry_count = died_retry_count;
@@ -345,7 +366,7 @@ void trans_HALF_DEAD(void)
 	/* If we time out, then we're going to stop. We need to reroute
          *  just in case the routes got changed before we timed out.
 	 */
-	syslog(LOG_INFO,"pppd restart timed out. Killing pppd.");
+	mon_syslog(LOG_NOTICE,"pppd restart timed out. Killing pppd.");
 	control_reroute();
     }
 }
@@ -354,59 +375,61 @@ void act_DISCONNECT(void)
 {
     dial_status = 0;
     if (disconnector) {
-        if (mode != MODE_DEV) {
-	    reopen_modem();
-            fork_dialer(disconnector, modem_fd);
-	}
-	else
-	{
-	/* added by jpd so that we can use a disconnect script to bring */
-	/* down the ipppd connection */
-	    fork_connect(disconnector);
-	}
+	reopen_modem();
+        dial_pid = run_shell(SHELL_NOWAIT,
+			"disconnector", disconnector, modem_fd);
     }
 }
 void trans_DISCONNECT(void)
 {
     if (dial_pid == 0) {
 	if (dial_status != 0)
-	    syslog(LOG_INFO,"Disconnect script failed");
+	    mon_syslog(LOG_WARNING,"Disconnect script failed");
 	GOTO(STATE_CLOSE);
     } else if (state_timeout == 0) {
-	syslog(LOG_INFO,"Disconnect script timed out. Killing script.");
+	mon_syslog(LOG_NOTICE,"Disconnect script timed out. Killing script.");
     }
 }
 
 void act_CLOSE(void)
 {
-    if (acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
+    if (call_timer_running
+    && acctlog && (acctfp = fopen(acctlog,"a")) != NULL) {
 	int duration = -call_start_time;
-	char *tm;
-	tm = cdate();
+	char *tm = cdate(time(&call_start_time));
 	duration += call_start_time;
+	call_timer_running = 0;
+
+	mon_syslog(LOG_NOTICE,
+	    "Disconnected. Call duration %d seconds.\n",
+	    duration);
         fprintf(acctfp,"%s: Disconnected. Call duration %d seconds.\n",
-	    tm,duration);
+	    tm, duration);
         fprintf(acctfp,"      IP transmitted %d bytes and received %d bytes.\n",
-	    txtotal,rxtotal);
+	    txtotal, rxtotal);
 	fclose(acctfp);
+	mon_syslog(LOG_NOTICE,
+	    "IP transmitted %d bytes and received %d bytes.\n",
+	    txtotal, rxtotal);
     }
     close_modem();
     interface_down();
     if (no_redial_delay == 0) {
 	if (redial_rtimeout == -1)
 	    redial_rtimeout = redial_timeout;
-        syslog(LOG_INFO,"Delaying %d seconds before clear to dial.",
+        mon_syslog(LOG_NOTICE,"Delaying %d seconds before clear to dial.",
 	    redial_rtimeout);
 	state_timeout = redial_rtimeout;
     } else if (no_redial_delay == 2) {
-        syslog(LOG_INFO,"Delaying %d seconds before clear to dial.",
+        mon_syslog(LOG_NOTICE,"Delaying %d seconds before clear to dial.",
 	    nodev_retry_timeout);
 	state_timeout = nodev_retry_timeout;
     }
 }
 void trans_CLOSE(void)
 {
-    if (request_up) GOTO(STATE_DOWN); /* STATE_DOWN handles the link-up-request */
+    if (request_up || request_down)
+	GOTO(STATE_DOWN); /* STATE_DOWN handles the link-up-request */
     if (no_redial_delay == 1) {
 	no_redial_delay = 0;
 	GOTO(STATE_RETRY);
@@ -419,7 +442,8 @@ void act_RETRY(void)
 }
 void trans_RETRY(void)
 {
-    if (request_up) GOTO(STATE_DOWN);
+    if (request_up || request_down)
+	GOTO(STATE_DOWN); /* STATE_DOWN handles the link-up-request */
     if (current_retry_count >= 0) GOTO(STATE_CONNECT);
     else GOTO(STATE_DOWN);
 }
@@ -427,7 +451,7 @@ void trans_RETRY(void)
 
 void act_ERROR(void)
 {
-    syslog(LOG_ERR,"Subprocess sent SIGKILL still alive after %d seconds. Reaping zombie.",kill_timeout);
+    mon_syslog(LOG_ERR,"Subprocess sent SIGKILL still alive after %d seconds. Reaping zombie.",kill_timeout);
     control_zombie();	/* deal with the zombie */
 }
 void trans_ERROR(void) {
@@ -436,7 +460,7 @@ void trans_ERROR(void) {
 
 void act_ZOMBIE(void)
 {
-    syslog(LOG_ERR,"Subprocess sent SIGKILL still alive after %d seconds. Reaping zombie.",kill_timeout);
+    mon_syslog(LOG_ERR,"Subprocess sent SIGKILL still alive after %d seconds. Reaping zombie.",kill_timeout);
 
     sig_chld(SIGKILL);	/* try to reap the zombie */
     if (dial_pid) {
@@ -480,6 +504,11 @@ struct {
 void output_state()
 {
     if (monitors) {
+	if (link_desc) {
+	    mon_write(MONITOR_VER2|MONITOR_STATE, "TITLE\n", 6);
+	    mon_write(MONITOR_VER2|MONITOR_STATE, link_desc, strlen(link_desc));
+	    mon_write(MONITOR_VER2|MONITOR_STATE, "\n", 1);
+	}
 	mon_write(MONITOR_STATE,"STATE\n",6);
 	mon_write(MONITOR_STATE,trans[state].name,strlen(trans[state].name));
 	mon_write(MONITOR_STATE,"\n",1);
@@ -491,7 +520,7 @@ void GOTO(int new_state)
     state_timeout = (trans[new_state].timeout)
                      ? *trans[new_state].timeout : -1;
     if (debug&DEBUG_STATE_CONTROL)
-    	syslog(LOG_DEBUG,"new state %s action %p timeout %d",
+    	mon_syslog(LOG_DEBUG,"new state %s action %p timeout %d",
 	    trans[new_state].name,trans[new_state].action,state_timeout);
     state = new_state;
     output_state();
