@@ -2,6 +2,7 @@
  * modem.c - Modem control functions.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  *
@@ -27,14 +28,17 @@
 
 #include "diald.h"
 
-static char *current_dev = 0;
+static char *lock_file = 0;
 static int rotate_offset = 0;
 
+char *current_dev = 0;
+int current_mode;
+int current_slip_encap;
+
 /* local variables */
 
 static struct termios inittermios;      /* Initial TTY termios */
 static int restore_term = 0;
-int pgrpid;
 
 #if B9600 == 9600
 /*
@@ -137,7 +141,7 @@ int translate_speed(int bps)
     for (speedp = speeds; speedp->speed_int; speedp++)
 	if (bps == speedp->speed_int)
 	    return speedp->speed_val;
-    syslog(LOG_WARNING, "speed %d not supported", bps);
+    mon_syslog(LOG_WARNING, "speed %d not supported", bps);
     return 0;
 }
 #endif
@@ -153,7 +157,7 @@ void set_up_tty(int fd, int local, int spd)
     struct termios tios;
 
     if (tcgetattr(fd, &tios) < 0) {
-	syslog(LOG_ERR, "could not get initial terminal attributes: %m");
+	mon_syslog(LOG_ERR, "could not get initial terminal attributes: %m");
     }
 
     tios.c_cflag = CS8 | CREAD | HUPCL;
@@ -176,7 +180,7 @@ void set_up_tty(int fd, int local, int spd)
     }
 
     if (tcsetattr(fd, TCSAFLUSH, &tios) < 0) {
-	syslog(LOG_ERR, "failed to set terminal attributes: %m");
+	mon_syslog(LOG_ERR, "failed to set terminal attributes: %m");
     }
 }
 
@@ -192,65 +196,13 @@ void setdtr(int fd, int on)
 }
 
 
-/*
- * fork_dialer - run a program to connect the serial device.
- */
-void fork_dialer(char *program, int fd)
-{
-    block_signals();
-
-    dial_pid = fork();
-
-    if (dial_pid < 0) {
-	unblock_signals();
-        syslog(LOG_ERR, "failed to fork dialer: %m");
-	/* FIXME: Probably this should not be fatal */
-        die(1);
-    }
-
-    if (dial_pid == 0) {
-        /* change the signal actions back to the defaults, then unblock them. */
-        default_sigacts();
-	unblock_signals();
-
-        /* make sure the child doesn't inherit any extra file descriptors */
-	close(proxy_mfd);      /* close the master pty endpoint */
-	close(proxy_sfd);      /* close the slave pty endpoint */
-	if (fifo_fd != -1) close(fifo_fd);
-	if (monitors) {
-	    MONITORS *c = monitors;
-	    while (c) {
-	    	close(c->fd);
-		c = c->next;
-	    }
-	}
-
-	/* make sure the stdin and stdout get directed to the modem */
-        if (fd != 0) { dup2(fd, 0); close(fd); }
-        dup2(0, 1);
-	/* FIXME: direct the stderr to the console? */
-        dup2(0, 2);
-
-	setenv("MODEM",current_dev,1);	/* set the current device */
-	if (fifoname)		/* set the current command FIFO (if any) */
-	    setenv("FIFO",fifoname,1);
-        execl("/bin/sh", "sh", "-c", program, (char *)0);
-        syslog(LOG_ERR, "could not exec /bin/sh: %m");
-        _exit(127);
-        /* NOTREACHED */
-    }
-
-    unblock_signals();
-    syslog(LOG_INFO,"Running connect (pid = %d).",dial_pid);
-}
-
 /*
  * Open up a modem and set up the desired parameters.
  */
 int open_modem()
 {
-    int npgrpid;
     int i;
+
     /*
      * Open the serial device and set it up.
      */
@@ -259,84 +211,107 @@ int open_modem()
     modem_fd = -1;
     dial_status = 0;
 
-    /* JPD changed mode_dev so that we can use a connect script to get */
-    /* ipppd to dial. This allows us to use a connector program to get */
-    /* ipppd to dial and make sure that ipppd successfully connects to */
-    /* the ISP. Otherwise, we can't be sure that ipppd successfully */
-    /* connected and diald would show that we we're connected */
-    if (mode == MODE_DEV) 
-    {
-        fork_connect(connector);
-        return 0;
-    }
-
+    /* If this is a request for diald to take over management of an
+     * interface that we didn't bring up ourselves there is little
+     * to do here except note what we are taking on.
+     */
     if (req_pid) {
-	/* The user has specified a device. Use it, no search or lock needed. */
-	if ((modem_fd = open(req_dev, O_RDWR | O_NDELAY)) < 0) {
-	    syslog(LOG_ERR,"Can't open requested device '%s'",req_dev);
-	    killpg(req_pid,SIGKILL);
-	    kill(req_pid,SIGKILL);
-	    req_pid = 0;
+	current_dev = req_dev;
+	use_req=1;
+
+	/* FIXME: If we are locking devices and the requested device
+	 * is not already locked we should probably lock it to prevent
+	 * anyone else trying it. This is a particular problem with
+	 * non-modem devices such as isdn? and ippp?.
+	 *   Possibly we should also inherit existing locks?
+	 */
+
+	modem_fd = open(current_dev[0] == '/' ? current_dev : "/dev/null",
+			O_RDWR | O_NDELAY);
+	if (modem_fd < 0) {
+	    mon_syslog(LOG_ERR, "Can't open requested device '%s'", req_dev);
+	    if (current_dev[0] != '/')
+		req_pid = 0;
+	    else {
+	    	killpg(req_pid, SIGKILL);
+	    	kill(req_pid, SIGKILL);
+	    }
 	    dial_status = -1;
 	    return 1;
 	}
-	current_dev = req_dev;
-	use_req=1;
+	if (lock_dev) lock_file = lock(current_dev);
     } else {
 	for (i = 0; i < device_count; i++) {
 	    current_dev = devices[(i+rotate_offset)%device_count];
-	    /*
-	     * March down the device list till we manage to open one up or
-	     * we run out of devices to try.
-	     */
 
-	    if (lock_dev && lock(current_dev) < 0)
+	    if (lock_dev && !(lock_file = lock(current_dev)))
 		continue;
 
 	    /* OK. Locked one, try to open it */
-	    if ((modem_fd = open(current_dev, O_RDWR | O_NDELAY)) >= 0)
+	    modem_fd = open(current_dev[0] == '/' ? current_dev : "/dev/null",
+			    O_RDWR | O_NDELAY);
+	    if (modem_fd >= 0)
 		break;
-	    else {
-	       syslog(LOG_ERR,"Error opening device %s: %m",current_dev);
-	    }
-	    current_dev = 0;
 
 	    /* That didn't work, get rid of the lock */
-	    if (lock_dev) unlock();
+	    mon_syslog(LOG_ERR,"Error opening device %s: %m",current_dev);
+	    if (lock_dev) unlock(lock_file);
 	}
 	if (modem_fd < 0) {
-	    syslog(LOG_INFO,"Couldn't find a free device to call out on.");
+	    mon_syslog(LOG_WARNING,"No devices free to call out on.");
+	    current_dev = 0;
 	    dial_status = -1;
 	    return 2;
 	}
+
+	if (rotate_devices)
+	    rotate_offset = (rotate_offset+1)%device_count;
+
+	current_dev = strdup(current_dev);
     }
 
-    if (rotate_devices)
-    	rotate_offset = (rotate_offset+1)%device_count;
+    if (current_dev[0] != '/')
+	current_mode = MODE_DEV;
+    else {
+	current_mode = mode;
+	current_slip_encap = slip_encap;
+    }
+
+    if (current_mode == MODE_DEV) {
+	/* If we have a connector we may need to run it to set up the
+	 * real interface.
+	 */
+	if (!req_pid && connector)
+		dial_pid = run_shell(SHELL_NOWAIT,
+				"connector", connector, modem_fd);
+	req_pid = 0;
+	return 0;
+    }
 
-    /* Make sure we are the session leader */
-    if ((npgrpid = setsid()) >= 0)
-	pgrpid = npgrpid;
 
     /* set device to be controlling tty */
-    /* This should only happen in SLIP mode */
-    if (mode == MODE_SLIP) {
-	if (ioctl(modem_fd, TIOCSCTTY, 1) < 0) {
-	    syslog(LOG_ERR, "failed to set modem to controlling tty: %m");
-	    die(1);
-	}
+    /* FIXME: we should not die here. If we go round again we may
+     * be able to use the next device in the list.
+     */
+    /* If this is outgoing this should have become our controlling
+     * tty when we opened it. If it is incoming it belongs to
+     * another session so we have to steal it.
+     */
+    if (ioctl(modem_fd, TIOCSCTTY, 1) < 0) {
+	mon_syslog(LOG_ERR, "failed to set modem to controlling tty: %m");
+	die(1);
+    }
 
-        if (tcsetpgrp(modem_fd, pgrpid) < 0) {
-	    syslog(LOG_ERR, "failed to set process group: %m");
-	    die(1);
-	}
+    if (tcsetpgrp(modem_fd, getpgrp()) < 0) {
+	mon_syslog(LOG_ERR, "open: failed to set tty process group: %m");
+	die(1);
     }
 
     /* Get rid of any initial line noise */
     tcflush(modem_fd, TCIOFLUSH);
 
     if (tcgetattr(modem_fd, &inittermios) < 0) {
-	syslog(LOG_ERR, "failed to get initial modem terminal attributes: %m");
+	mon_syslog(LOG_ERR, "failed to get initial modem terminal attributes: %m");
     }
 
     /* So we don't try to restore if we die before this */
@@ -344,17 +319,24 @@ int open_modem()
 
     /* Clear the NDELAY flag now */
     if (fcntl(modem_fd,F_SETFL,fcntl(modem_fd,F_GETFL)&~(O_NDELAY)) < 0)
-	syslog(LOG_ERR, "failed to clear O_NDELAY flag: %m"), die(1);
+	mon_syslog(LOG_ERR, "failed to clear O_NDELAY flag: %m"), die(1);
 
     if (!req_pid) {
+	int line_disc;
+
 	/* hang up and then start again */
+	set_up_tty(modem_fd, 1, inspeed);
+	if (ioctl(modem_fd, TIOCGETD, &line_disc) < 0 || line_disc != N_TTY) {
+	    line_disc = N_TTY;
+	    ioctl(modem_fd, TIOCSETD, &line_disc);
+	}
 	set_up_tty(modem_fd, 1, 0);
 	sleep(1);
 	set_up_tty(modem_fd, 1, inspeed);
 
 	/* Get rid of any initial line noise after the hangup */
 	tcflush(modem_fd, TCIOFLUSH);
-	fork_dialer(connector, modem_fd);
+	dial_pid = run_shell(SHELL_NOWAIT, "connector", connector, modem_fd);
     } else {
 	/* someone else opened the line, we just set the mode */
 	set_up_tty(modem_fd, 0, inspeed);
@@ -367,45 +349,45 @@ int open_modem()
  */
 void reopen_modem()
 {
-    int npgrpid;
+    if (current_mode == MODE_DEV)
+	return;
 
     if(debug&DEBUG_VERBOSE)
-	syslog(LOG_INFO,"Reopening modem device");
+	mon_syslog(LOG_INFO,"Reopening modem device");
 
     close(modem_fd);
     sleep(1);
     if ((modem_fd = open(current_dev, O_RDWR | O_NDELAY)) < 0) {
-	syslog(LOG_ERR,"Can't reopen device '%s'",current_dev);
+	mon_syslog(LOG_ERR,"Can't reopen device '%s'",current_dev);
     } else {
-	/* Make sure we are the session leader */
-	if ((npgrpid = setsid()) >= 0)
-	    pgrpid = npgrpid;
-
 	/* set device to be controlling tty */
-	/* This should only happen in SLIP mode */
-	if (mode == MODE_SLIP) {
-	    if (ioctl(modem_fd, TIOCSCTTY, 1) < 0) {
-		syslog(LOG_ERR, "failed to set modem to controlling tty: %m");
-		die(1);
-	    }
+	if (ioctl(modem_fd, TIOCSCTTY, 1) < 0) {
+	    mon_syslog(LOG_ERR, "reopen: failed to set modem to controlling tty: %m");
+	    die(1);
+	}
 
-	    if (tcsetpgrp(modem_fd, pgrpid) < 0) {
-		syslog(LOG_ERR, "failed to set process group: %m");
-		die(1);
-	    }
+	if (tcsetpgrp(modem_fd, getpgrp()) < 0) {
+	    mon_syslog(LOG_ERR, "reopen: failed to set process group: %m");
+	    die(1);
 	}
 
 	set_up_tty(modem_fd, 1, inspeed);
 	/* Clear the NDELAY flag now */
 	if (fcntl(modem_fd,F_SETFL,fcntl(modem_fd,F_GETFL)&~(O_NDELAY)) < 0)
-	    syslog(LOG_ERR, "failed to clear O_NDELAY flag: %m"), die(1);
+	    mon_syslog(LOG_ERR, "failed to clear O_NDELAY flag: %m"), die(1);
     }
 }
 
 void finish_dial()
 {
-    if (!req_pid)
+    if (current_mode == MODE_DEV)
+	return;
+
+    if (!req_pid) {
+	if (tcsetpgrp(modem_fd, getpgrp()) < 0)
+	    mon_syslog(LOG_ERR, "finish: failed to set pgrp: %m");
         set_up_tty(modem_fd, 0, inspeed);
+    }
 }
 
 /*
@@ -413,47 +395,54 @@ void finish_dial()
  */
 void close_modem()
 {
-    if (mode == MODE_DEV) {
-	req_pid = 0;
-	modem_fd = -1 ;
-	return ;
-    }
+    if (current_dev && debug&DEBUG_VERBOSE)
+        mon_syslog(LOG_INFO,"Closing %s", current_dev);
 
-    if (debug&DEBUG_VERBOSE)
-        syslog(LOG_INFO,"Closing modem line.");
-
-    if (modem_fd < 0) {
+    if (current_dev)
+	free(current_dev);
+    current_dev = 0;
+    if (modem_fd < 0)
  	return;
+
+    if (current_mode != MODE_DEV) {
+	tcsetpgrp(modem_fd, getpgrp());
+
+	/* Get rid of what ever might be waiting to go out still */
+	tcflush(modem_fd, TCIOFLUSH);
+
+	/*
+	 * Restore the initial termio settings.
+	 */
+
+	if (restore_term) {
+	    tcsetattr(modem_fd, TCSANOW, &inittermios);
+	}
+
+	/*
+	 * Hang up the modem up by dropping the DTR.
+	 * We do this because the initial termio settings
+	 * may not have set HUPCL. This forces the issue.
+	 * We need the sleep to give the modem a chance to hang
+	 * up before we get another program asserting the DTR.
+	 */
+	setdtr(modem_fd, 0);
+	sleep(1);
     }
 
-    /* Get rid of what ever might be waiting to go out still */
-    tcflush(modem_fd, TCIOFLUSH);
-
-    /*
-     * Restore the initial termio settings.
-     */
-
-    if (restore_term) {
-	tcsetattr(modem_fd, TCSANOW, &inittermios);
-    }
-
-    /*
-     * Hang up the modem up by dropping the DTR.
-     * We do this because the initial termio settings
-     * may not have set HUPCL. This forces the issue.
-     * We need the sleep to give the modem a chance to hang
-     * up before we get another program asserting the DTR.
-     */
-    setdtr(modem_fd, 0);
-    sleep(1);
-
     close(modem_fd);
-    if (use_req) {
-	if (debug&DEBUG_VERBOSE)
-	    syslog(LOG_INFO, "Killing requesting shell pid %d",req_pid);
-	killpg(req_pid, SIGKILL);
-	kill(req_pid, SIGKILL);
-	req_pid = 0;
-    } else if (lock_dev) unlock();
     modem_fd = -1;
+
+    if (use_req) {
+	if (req_pid) {
+	    if (debug&DEBUG_VERBOSE)
+		mon_syslog(LOG_INFO, "Killing requesting shell pid %d",req_pid);
+	    killpg(req_pid, SIGKILL);
+	    kill(req_pid, SIGKILL);
+	    req_pid = 0;
+	}
+    }
+    if (lock_file) {
+	unlock(lock_file);
+	lock_file = NULL;
+    }
 }
