@@ -2,13 +2,14 @@
  * ppp.c - ppp and pppd control.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
 
 #include "diald.h"
 
-#ifdef 0
+#if 0
 #ifdef PPP_VERSION_2_2_0
 #include <linux/ppp_defs.h>
 #include <linux/if_ppp.h>
@@ -37,14 +38,14 @@ void ppp_start()
 
     if (link_pid < 0) {
         unblock_signals();
-	syslog(LOG_ERR, "failed to fork pppd: %m");
+	mon_syslog(LOG_ERR, "failed to fork pppd: %m");
 	die(1);
     }
 
 #define ADD_ARG(arg) { argv[i] = arg; argv_len += strlen(argv[i++]) + 1; }
     
     if (link_pid == 0) {
-	char **argv = (char **)malloc(sizeof(char *)*(pppd_argc+11));
+	char **argv = (char **)malloc(sizeof(char *)*(pppd_argc+12));
 	int argv_len = 0;
 	char buf[24], *argv_buf;
 	int i = 0, j;;
@@ -53,6 +54,7 @@ void ppp_start()
         unblock_signals();
 
 	ADD_ARG(path_pppd);
+	ADD_ARG("-defaultroute");
 	ADD_ARG("-detach");
 	if (modem) ADD_ARG("modem");
 	if (crtscts) ADD_ARG("crtscts");
@@ -78,7 +80,7 @@ void ppp_start()
 	    strcat(argv_buf, argv[i]);
 	    strcat(argv_buf, " ");
 	  }
-	  syslog(LOG_DEBUG, "Running pppd: %s", argv_buf);
+	  mon_syslog(LOG_DEBUG, "Running pppd: %s", argv_buf);
 	}
 
 	/* make sure pppd is the session leader and has the controlling
@@ -96,12 +98,12 @@ void ppp_start()
 
 	execv(path_pppd,argv);
 
-	syslog(LOG_ERR, "could not exec %s: %m",path_pppd);
+	mon_syslog(LOG_ERR, "could not exec %s: %m",path_pppd);
 	_exit(99);
 	/* NOTREACHED */
     }
     unblock_signals();
-    syslog(LOG_INFO,"Running pppd (pid = %d).",link_pid);
+    mon_syslog(LOG_INFO,"Running pppd (pid = %d).",link_pid);
 }
 
 /*
@@ -121,7 +123,12 @@ void ppp_start()
 
 int ppp_set_addrs()
 {
-    ulong laddr = 0, raddr = 0;
+    static int sock = -1;
+    ulong laddr = 0, raddr = 0, baddr = 0;
+
+    /* We need a socket. Any socket... */
+    if (sock < 0)
+	sock = socket(AF_INET, SOCK_DGRAM, 0);
 
     /* Try to get the interface number if we don't know it yet. */
     if (link_iface == -1) {
@@ -140,8 +147,8 @@ int ppp_set_addrs()
 	SET_SA_FAMILY (ifr.ifr_dstaddr, AF_INET); 
 	SET_SA_FAMILY (ifr.ifr_netmask, AF_INET); 
 	sprintf(ifr.ifr_name,"ppp%d",link_iface);
-	if (ioctl(snoopfd, SIOCGIFFLAGS, (caddr_t) &ifr) == -1) {
-	   syslog(LOG_ERR,"failed to read ppp interface status");
+	if (ioctl(sock, SIOCGIFFLAGS, (caddr_t) &ifr) == -1) {
+	   mon_syslog(LOG_ERR,"failed to read ppp interface status: %m");
 	   return 0;
 	}
 	if (!(ifr.ifr_flags & IFF_UP))
@@ -156,61 +163,73 @@ int ppp_set_addrs()
 	}
 
 	/* Ok, the interface is up, grab the addresses. */
-	if (ioctl(snoopfd, SIOCGIFADDR, (caddr_t) &ifr) == -1)
-		syslog(LOG_ERR,"failed to get ppp local address: %m");
+	if (ioctl(sock, SIOCGIFADDR, (caddr_t) &ifr) == -1)
+		mon_syslog(LOG_ERR,"failed to get ppp local address: %m");
 	else
        	    laddr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr;
 
-	if (ioctl(snoopfd, SIOCGIFDSTADDR, (caddr_t) &ifr) == -1) 
-	   syslog(LOG_ERR,"failed to get ppp remote address: %m");
+	if (ioctl(sock, SIOCGIFDSTADDR, (caddr_t) &ifr) == -1) 
+	   mon_syslog(LOG_ERR,"failed to get ppp remote address: %m");
 	else
 	   raddr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr;
 
+	if (ioctl(sock, SIOCGIFBRDADDR, (caddr_t) &ifr) == -1) 
+	   mon_syslog(LOG_ERR,"failed to get ppp broadcast address: %m");
+	else
+	   baddr = ((struct sockaddr_in *) &ifr.ifr_addr)->sin_addr.s_addr;
+
  	/* Check the MTU, see if it matches what we asked for. If it
 	 * doesn't warn the user and adjust the MTU setting.
 	 * (NOTE: Adjusting the MTU setting may cause kernel nastyness...)
 	 */
-	if (ioctl(snoopfd, SIOCGIFMTU, (caddr_t) &ifr) == -1) {
-	    syslog(LOG_ERR,"failed to get ppp mtu setting: %m");
+	if (ioctl(sock, SIOCGIFMTU, (caddr_t) &ifr) == -1) {
+	    mon_syslog(LOG_ERR,"failed to get ppp mtu setting: %m");
 	} else {
 	    if (ifr.ifr_mtu != mtu) {
-	        syslog(LOG_ERR,"PPP negotiated mtu of %d does not match requested setting %d.",ifr.ifr_mtu,mtu);
-		syslog(LOG_ERR,"Attempting to auto adjust mtu.");
-		syslog(LOG_ERR,"Restart diald with mtu set to %d to avoid errors.",ifr.ifr_mtu);
+	        mon_syslog(LOG_WARNING,"PPP negotiated mtu of %d does not match requested setting %d.",ifr.ifr_mtu,mtu);
+		mon_syslog(LOG_WARNING,"Attempting to auto adjust mtu.");
+		mon_syslog(LOG_WARNING,"Restart diald with mtu set to %d to avoid errors.",ifr.ifr_mtu);
 		mtu = ifr.ifr_mtu;
-	        proxy_config(orig_local_ip,orig_remote_ip);
 	    }
 	}
 
 	if (dynamic_addrs) {
 	    /* only do the configuration in dynamic mode. */
 	    struct in_addr addr;
+	    addr.s_addr = baddr;
+	    if (broadcast_ip) free(broadcast_ip);
+	    broadcast_ip = strdup(inet_ntoa(addr));
 	    addr.s_addr = raddr;
+	    if (remote_ip) free(remote_ip);
 	    remote_ip = strdup(inet_ntoa(addr));
 	    addr.s_addr = laddr;
+	    if (local_ip) free(local_ip);
 	    local_ip = strdup(inet_ntoa(addr));
 	    local_addr = laddr;
-	    syslog(LOG_INFO,"New addresses: local %s, remote %s.",
-		local_ip,remote_ip);
-	    /* have to reset the proxy if we won't be rerouting... */
-	    if (!do_reroute) {
-		/* If we are rerouting, then we have a window without
-		 * routes here. The proxy_config calls ifconfig, which
-		 * clobbers all the routes.
-		 */
-	    	proxy_config(local_ip,remote_ip);
-    		set_ptp("sl",proxy_iface,remote_ip,1);
-	        add_routes("sl",proxy_iface,local_ip,remote_ip,1);
+	    if (dynamic_addrs > 1) {
+		if (orig_broadcast_ip) free(orig_broadcast_ip);
+		orig_broadcast_ip = strdup(broadcast_ip);
+		if (orig_remote_ip) free(orig_remote_ip);
+		orig_remote_ip = strdup(remote_ip);
+		if (orig_local_ip) free(orig_local_ip);
+		orig_local_ip = strdup(local_ip);
 	    }
+	    mon_syslog(LOG_INFO, "New addresses: local %s%s%s%s%s",
+		local_ip,
+		remote_ip ? ", remote " : "",
+		remote_ip ? remote_ip : "",
+		broadcast_ip ? ", broadcast " : "",
+		broadcast_ip ? broadcast_ip : "");
 	}
 
-	/* This is redundant in normal operation, but if we
-	 * have to restart the link, then this is necessary...
+	/* The pppd should have configured the interface but there
+	 * may be user or default routes to add :-(.
 	 */
-	set_ptp("ppp",link_iface,remote_ip,0);
+	iface_start("link", "ppp", link_iface,
+	    local_ip, remote_ip, broadcast_ip);
+	if (proxy.stop)
+	    proxy.stop(&proxy);
 
-	if (do_reroute)
-	     add_routes("ppp",link_iface,local_ip,remote_ip,0);
 	return 1;
     }
     return 0;
@@ -218,10 +237,6 @@ int ppp_set_addrs()
 
 int ppp_dead()
 {
-    /* pppd is way too enthusiastic about deleting routes it had
-     * nothing to do with creating. Therefore, we have to reestablish
-     * the proxy routes here, including the point to point route!
-     */
     if (link_pid == 0)
 	ppp_reroute();
     return (link_pid == 0);
@@ -235,7 +250,7 @@ int ppp_route_exists()
     FILE *fp;
     sprintf(buf,"%s -n",path_route);
     if ((fp = popen(buf,"r"))==NULL) {
-        syslog(LOG_ERR,"Could not run command '%s': %m",buf);
+        mon_syslog(LOG_ERR,"Could not run command '%s': %m",buf);
 	return 0;	/* assume half dead in this case... */
     }
 
@@ -255,7 +270,7 @@ int ppp_rx_count()
     FILE *fp;
     sprintf(buf,"%s ppp%d",path_ifconfig,link_iface);
     if ((fp = popen(buf,"r"))==NULL) {
-        syslog(LOG_ERR,"Could not run command '%s': %m",buf);
+        mon_syslog(LOG_ERR,"Could not run command '%s': %m",buf);
 	return 0;	/* assume half dead in this case... */
     }
 
@@ -277,20 +292,14 @@ void ppp_stop()
 
 void ppp_reroute()
 {
-    /* Restore the original proxy routing */
-    /* If there was a change in the proxy addresses
-     * (i.e., we are running with -reroute), then
-     * this will change introduce a window in which
-     * we loose our routes. There does not seem to be a way
-     * to avoid this. Sigh.
-     */
-    proxy_config(orig_local_ip,orig_remote_ip);
-    set_ptp("sl",proxy_iface,orig_remote_ip,1);
-    add_routes("sl",proxy_iface,orig_local_ip,orig_remote_ip,1);
-    local_addr = inet_addr(orig_local_ip);
-    /* If we did routing on the ppp link, remove it */
-    if (do_reroute && link_iface != -1)
-    	del_routes("ppp",link_iface,local_ip,remote_ip,0);
+    /* Restore the original proxy. */
+    if (proxy.start && (!blocked || blocked_route))
+	proxy.start(&proxy);
+    local_addr = (orig_local_ip ? inet_addr(orig_local_ip) : 0);
+
+    if (link_iface != -1)
+    	iface_stop("link", "ppp", link_iface,
+	    local_ip, remote_ip, broadcast_ip);
     link_iface = -1;
 }
 
