@@ -2,6 +2,7 @@
  * slip.c - Slip specific code in diald.
  *
  * Copyright (c) 1994, 1995, 1996 Eric Schenk.
+ * Copyright (c) 1999 Mike Jagdis.
  * All rights reserved. Please see the file LICENSE which should be
  * distributed with this software for terms of use.
  */
@@ -31,10 +32,11 @@ static void start_bootp()
      * write my own popen() call the same way I have my open system() call.
      */
     if ((bootpfp = popen(buf,"r"))==NULL) {
-	syslog(LOG_ERR,"Could not run command '%s': %m",buf);
+	mon_syslog(LOG_ERR,"Could not run command '%s': %m",buf);
 	die(1);
     }
-    pipe_init(fileno(bootpfp),&bootp_pipe);
+    pipe_init("BOOTP", (ACCESS_CONTROL|ACCESS_MESSAGE|ACCESS_DYNAMIC),
+	fileno(bootpfp),&bootp_pipe, 0);
     have_local = 0;
     have_remote = 0;
     waiting_for_bootp = 1;
@@ -52,11 +54,11 @@ static int grab_addr(char **var)
     while (1) {
 	if ((len = read(modem_fd,&c,1)) < 0) {
 	    if (fail) {
-		syslog(LOG_ERR,"Timeout waiting for dynamic slip addresses");
+		mon_syslog(LOG_ERR,"Timeout waiting for dynamic slip addresses");
 		return 0;
 	    }
 	    if (errno != EINTR || terminate) {
-	        syslog(LOG_ERR,"Error reading from modem: %m");
+	        mon_syslog(LOG_ERR,"Error reading from modem: %m");
 	        return 0;
             }
         }
@@ -80,12 +82,13 @@ static int grab_addr(char **var)
 		break;
 	    }
 	    if (i >= 128) {
-		syslog(LOG_ERR,"Buffer overflow when reading IP address"),
+		mon_syslog(LOG_ERR,"Buffer overflow when reading IP address"),
 		die(1);
 	    }
 	}
     }
 done:
+    if (*var) free(*var);
     *var = strdup(buffer);
     return 1;
 }
@@ -106,15 +109,13 @@ struct timer_lst failt;
 
 void slip_start(void)
 {
-    char buf[128];
     int disc, sencap;
-    int res;
 
     rx_count = -1;
 
     if (dynamic_addrs && !force_dynamic) {
 	if (debug&DEBUG_VERBOSE)
-	    syslog(LOG_INFO,"Fetching IP addresses from SLIP server");
+	    mon_syslog(LOG_DEBUG,"Fetching IP addresses from SLIP server");
 	if (dynamic_mode != DMODE_BOOTP) {
 	    fail = 0;
 	    failt.data = 0;
@@ -128,67 +129,63 @@ void slip_start(void)
 	    if (dynamic_mode == DMODE_LOCAL_REMOTE)
 		if (!grab_addr(&remote_ip)) return;
 	    del_timer(&failt);
-	    syslog(LOG_INFO,"New addresses: local %s, remote %s.",
+	    if (dynamic_addrs > 1) {
+		if (orig_remote_ip) free(orig_remote_ip);
+		orig_remote_ip = strdup(remote_ip);
+		if (orig_local_ip) free(orig_local_ip);
+		orig_local_ip = strdup(local_ip);
+	    }
+	    mon_syslog(LOG_INFO,"New addresses: local %s, remote %s.",
 		local_ip,remote_ip);
 	}
     }
 
     if (ioctl(modem_fd, TIOCGETD, &start_disc) < 0)
-	syslog(LOG_ERR,"Can't get line discipline on proxy device: %m"), die(1);
+	mon_syslog(LOG_ERR,"Can't get line discipline on proxy device: %m"), die(1);
 
     /* change line disciple to SLIP and set the SLIP encapsulation */
     disc = N_SLIP;
     if ((link_iface = ioctl(modem_fd, TIOCSETD, &disc)) < 0) {
 	if (errno == ENFILE) {
-	   syslog(LOG_ERR,"No free slip device available for link."), die(1);
+	   mon_syslog(LOG_ERR,"No free slip device available for link."), die(1);
 	} else if (errno == EEXIST) {
-	    syslog(LOG_ERR,"Link device already in slip mode!?");
+	    mon_syslog(LOG_ERR,"Link device already in slip mode!?");
 	} else if (errno == EINVAL) {
-	    syslog(LOG_ERR,"SLIP not supported by kernel, can't build link.");
+	    mon_syslog(LOG_ERR,"SLIP not supported by kernel, can't build link.");
 	    die(1);
 	} else
-	   syslog(LOG_ERR,"Can't set line discipline: %m"), die(1);
+	   mon_syslog(LOG_ERR,"Can't set line discipline: %m"), die(1);
     }
 
-    if (ioctl(modem_fd, SIOCSIFENCAP, &slip_encap) < 0)
-	syslog(LOG_ERR,"Can't set encapsulation: %m"), die(1);
+    if (ioctl(modem_fd, SIOCSIFENCAP, &current_slip_encap) < 0)
+	mon_syslog(LOG_ERR,"Can't set encapsulation: %m"), die(1);
 
 #ifdef SIOCSKEEPALIVE
     if (keepalive && (ioctl(modem_fd, SIOCSKEEPALIVE, &keepalive) < 0))
-      syslog(LOG_ERR, "Can't set keepalive: %m (ignoring error)");
+      mon_syslog(LOG_ERR, "Can't set keepalive: %m (ignoring error)");
 #endif
 
 #ifdef SIOCSOUTFILL
     if (outfill && (ioctl(modem_fd, SIOCSOUTFILL, &outfill) < 0))
-      syslog(LOG_ERR, "Can't set outfill: %m (ignoring error)");
+      mon_syslog(LOG_ERR, "Can't set outfill: %m (ignoring error)");
 #endif
 
     /* verify that it worked */
     if (ioctl(modem_fd, TIOCGETD, &disc) < 0)
-	syslog(LOG_ERR,"Can't get line discipline: %m"), die(1);
+	mon_syslog(LOG_ERR,"Can't get line discipline: %m"), die(1);
     if (ioctl(modem_fd, SIOCGIFENCAP, &sencap) < 0)
-	syslog(LOG_ERR,"Can't get encapsulation: %m"), die(1);
+	mon_syslog(LOG_ERR,"Can't get encapsulation: %m"), die(1);
 
-    if (disc != N_SLIP || sencap != slip_encap)
-        syslog(LOG_ERR,"Couldn't set up the slip link correctly!"), die(1);
+    if (disc != N_SLIP || sencap != current_slip_encap)
+        mon_syslog(LOG_ERR,"Couldn't set up the slip link correctly!"), die(1);
 
     if (debug&DEBUG_VERBOSE)
-        syslog(LOG_INFO,"Slip link established on interface sl%d",
+        mon_syslog(LOG_INFO,"Slip link established on interface sl%d",
 	    link_iface);
 
     /* mark the interface as up */
-    if (netmask) {
-        sprintf(buf,"%s sl%d %s pointopoint %s netmask %s mtu %d up",
-	    path_ifconfig,link_iface,local_ip,remote_ip,netmask,mtu);
-    } else {
-        sprintf(buf,"%s sl%d %s pointopoint %s mtu %d up",
-	    path_ifconfig,link_iface,local_ip,remote_ip,mtu);
-    }
-    res = system(buf);
-    report_system_result(res,buf);
-
-    /* Set the routing for the new slip interface */
-    set_ptp("sl",link_iface,remote_ip,0);
+    iface_start("link", "sl", link_iface,
+	local_ip, remote_ip, broadcast_ip);
 
     /* run bootp if it is asked for */
     if (dynamic_addrs && dynamic_mode == DMODE_BOOTP && !force_dynamic) start_bootp();
@@ -198,10 +195,8 @@ void slip_start(void)
 
 int slip_set_addrs()
 {
-    char buf[128];
     char type[30];
     char addr[30];
-    int res;
 
     if (waiting_for_bootp && (!have_local || !have_remote)) {
 	int i;
@@ -216,9 +211,11 @@ int slip_set_addrs()
 		if (sscanf(buf,"%[^=]=%s\n",type,addr) == 2) {
 		    if (strcmp(type,"SERVER")==0) {
 			have_remote = 1;
+			if (remote_ip) free(remote_ip);
 			remote_ip = strdup(addr);
 		    } else if (strcmp(type,"IPADDR")==0) {
 			have_local = 1;
+			if (local_ip) free(local_ip);
 			local_ip = strdup(addr);
 		    }
 		}
@@ -234,7 +231,13 @@ int slip_set_addrs()
 
     if (waiting_for_bootp) {
         pclose(bootpfp);
-	syslog(LOG_INFO,"New addresses: local %s, remote %s.",
+	if (dynamic_addrs > 1) {
+	    if (orig_remote_ip) free(orig_remote_ip);
+	    orig_remote_ip = strdup(remote_ip);
+	    if (orig_local_ip) free(orig_local_ip);
+	    orig_local_ip = strdup(local_ip);
+	}
+	mon_syslog(LOG_INFO,"New addresses: local %s, remote %s.",
 	    local_ip,remote_ip);
     	waiting_for_bootp = 0;
     }
@@ -247,32 +250,14 @@ int slip_set_addrs()
 	if (slip_rx_count() == rx_count) return 0;
     }
 
-    /* redo the interface marking and the routing since BOOTP will change it */
-    if (netmask) {
-        sprintf(buf,"%s sl%d %s pointopoint %s netmask %s mtu %d up",
-	    path_ifconfig,link_iface,local_ip,remote_ip,netmask,mtu);
-    } else {
-        sprintf(buf,"%s sl%d %s pointopoint %s mtu %d up",
-	    path_ifconfig,link_iface,local_ip,remote_ip,mtu);
-    }
-    res = system(buf);
-    report_system_result(res,buf);
-
-    /* Set the routing for the new slip interface */
-    set_ptp("sl",link_iface,remote_ip,0);
-
-    if (dynamic_addrs || force_dynamic) {
+    if ((dynamic_addrs || force_dynamic) && local_ip) {
 	local_addr = inet_addr(local_ip);
-	/* have to reset the proxy if we won't be rerouting... */
-	if (!do_reroute) {
-	    proxy_config(local_ip,remote_ip);
-    	    set_ptp("sl",proxy_iface,remote_ip,1);
-	    add_routes("sl",proxy_iface,local_ip,remote_ip,1);
-	}
     }
 
-    if (do_reroute)
-        add_routes("sl",link_iface,local_ip,remote_ip,0);
+    iface_start("link", "sl", link_iface,
+	local_ip, remote_ip, broadcast_ip);
+    if (proxy.stop)
+	proxy.stop(&proxy);
 
     return 1;
 }
@@ -296,14 +281,14 @@ void slip_stop()
 
 void slip_reroute()
 {
-    /* Restore the original proxy routing */
-    proxy_config(orig_local_ip,orig_remote_ip);
-    set_ptp("sl",proxy_iface,orig_remote_ip,1);
-    add_routes("sl",proxy_iface,orig_local_ip,orig_remote_ip,1);
-    local_addr = inet_addr(orig_local_ip);
-    /* If we did routing on the slip link, remove it */
-    if (do_reroute && link_iface != -1) /* just in case we get called twice */
-    	del_routes("sl",link_iface,local_ip,remote_ip,0);
+    /* Restore the original proxy. */
+    if (proxy.start && (!blocked || blocked_route))
+	proxy.start(&proxy);
+    local_addr = (orig_local_ip ? inet_addr(orig_local_ip) : 0);
+
+    if (link_iface != -1)
+    	iface_stop("link", "sl", link_iface,
+	    local_ip, remote_ip, broadcast_ip);
     link_iface = -1;
 }
 
@@ -314,7 +299,7 @@ int slip_rx_count()
     FILE *fp;
     sprintf(buf,"%s sl%d",path_ifconfig,link_iface);
     if ((fp = popen(buf,"r"))==NULL) {
-        syslog(LOG_ERR,"Could not run command '%s': %m",buf);
+        mon_syslog(LOG_ERR,"Could not run command '%s': %m",buf);
 	return 0;	/* assume half dead in this case... */
     }
 
