@@ -1,5 +1,18 @@
 #include "diald.h"
 
+/* Grumble. Attempt to generate a nicely formatted ascii date without
+ * a built in newline.
+ */
+char *cdate(time_t now)
+{
+    static char dt[128];
+    int len;
+
+    len = strftime(dt, 128, "%c %Z", localtime(&now));
+    if (len == 128) dt[len] = 0;
+    return dt;
+}
+
 int getsn(FILE *fp,char *buf,int len)
 {
     int c;
@@ -70,13 +83,16 @@ int getprotocol(const char *name)
 
 char *getprotonumber(int proto)
 {
+    static char buf[16];
     int i;
+
     if (!proto_init)
 	init_protos();
     for (i = 0; i < proto_lines; i++)
 	if (protos[i].proto == proto)
 	    return protos[i].name;
-    return 0;
+    sprintf(buf, "%d", proto);
+    return buf;
 }
 
 struct serv {
@@ -133,6 +149,7 @@ int getservice(const char *name, const char *proto)
     return 0;
 }
 
+#if 0
 /* Stuff needed because to keep checker happy,
  * because the fast versions of these address memory in four
  * byte blocks, which are often outside of permitted ranges.
@@ -169,3 +186,4 @@ char *strcat(char *dest, const char *src)
 	while ((*p++ = *src++));
 	return dest;
 }
+#endif
